#! /usr/bin/env python3
# SPDX-License-Identifier: GPL-2.0-or-later

# depthcharge-tools string utilities
# Copyright (C) 2022 Alper Nebi Yasak <alpernebiyasak@gmail.com>
# See COPYRIGHT and LICENSE files for full copyright information.

import ast
import re

def bytesize_suffixes():
    def long_forms(x):
        formats = ("{}", "{}byte", "{} byte", "{}bytes", "{} bytes")
        cases = (str.upper, str.lower, str.title)
        for f in formats:
            for c in cases:
                yield c(f.format(x))

    for size, suffixes in {
        1:       ("B", "byte", "bytes", ""),
        1e3:     ("kB", "KB", *long_forms("kilo")),
        1e6:     ("MB", *long_forms("mega")),
        1e9:     ("GB", *long_forms("giga")),
        1e12:    ("TB", *long_forms("tera")),
        1e15:    ("PB", *long_forms("peta")),
        1e18:    ("EB", *long_forms("exa")),
        1e21:    ("ZB", *long_forms("zetta")),
        1e24:    ("YB", *long_forms("yotta")),
        2 ** 10: ("kiB", "KiB", "K", *long_forms("kibi")),
        2 ** 20: ("MiB", "M", *long_forms("mebi")),
        2 ** 30: ("GiB", "G", *long_forms("gibi")),
        2 ** 40: ("TiB", "T", *long_forms("tebi")),
        2 ** 50: ("PiB", "P", *long_forms("pebi")),
        2 ** 60: ("EiB", "E", *long_forms("exbi")),
        2 ** 70: ("ZiB", "Z", *long_forms("zebi")),
        2 ** 80: ("YiB", "Y", *long_forms("yobi")),
    }.items():
        for suffix in suffixes:
            yield (suffix.strip(), int(size))

bytesize_suffixes = dict(bytesize_suffixes())


def parse_bytesize(val):
    if val is None:
        return None

    try:
        return int(val)
    except:
        pass

    try:
        return int(ast.literal_eval(val))
    except:
        pass

    try:
        s = str(val)
        suffix = re.search("[a-zA-Z\s]*\Z", s)[0].strip()
        number = s.rpartition(suffix)[0].strip()
        multiplier = bytesize_suffixes[suffix]
        return int(ast.literal_eval(number)) * multiplier

    except Exception as err:
        raise ValueError(
            "Cannot convert '{}' to a byte-size."
            .format(val)
        )



#! /usr/bin/env python3
# SPDX-License-Identifier: GPL-2.0-or-later

# depthcharge-tools collections utilities
# Copyright (C) 2021-2022 Alper Nebi Yasak <alpernebiyasak@gmail.com>
# See COPYRIGHT and LICENSE files for full copyright information.

import collections

# Inheritance for config sections
class ConfigDict(collections.OrderedDict):
    def __getitem__(self, key):
        super_ = super()
        if not isinstance(key, str) or "/" not in key:
            return super_.__getitem__(key)

        def getitem(key):
            try:
                return super_.__getitem__(key)
            except KeyError:
                return KeyError

        def parents(leaf):
            idx = leaf.find("/")
            while idx != -1:
                yield leaf[:idx]
                idx = leaf.find("/", idx + 1)
            yield leaf

        items = list(
            item for item in reversed([
                getitem(p) for p in parents(key)
            ]) if item != KeyError
        )

        if all(isinstance(i, dict) for i in items):
            return collections.ChainMap(*items)

        if items:
            return items[0]

        raise KeyError(key)


# To write config sections in sort order
def SortedDict(key=None):
    if not callable(key):
        raise TypeError(
            "SortedDict argument must be a callable, not {}"
            .format(type(key).__name__)
        )

    class SortedDict(collections.UserDict):
        __key = key

        def __iter__(self):
            yield from sorted(super().__iter__(), key=type(self).__key)

    return SortedDict


def TypedList(T):
    if not isinstance(T, type):
        raise TypeError(
            "TypedList argument must be a type, not {}"
            .format(type(T).__name__)
        )

    name = "TypedList.{}List".format(str.title(T.__name__))

    class TypedList(collections.UserList):
        __type = T
        __name__ = name
        __qualname__ = name

        def __init__(self, initlist=None):
            if initlist is not None:
                self.__typecheck(*initlist)
            super().__init__(initlist)

        def __typecheck(self, *values):
            if not all(isinstance(value, self.__type) for value in values):
                raise TypeError(
                    "{} items must be of type {}."
                    .format(type(self).__name__, self.__type.__name__)
                )

        def __setitem__(self, idx, value):
            self.__typecheck(value)
            return super().__setitem__(idx, value)

        def __iadd__(self, other):
            self.__typecheck(*other)
            return super().__iadd__(other)

        def append(self, value):
            self.__typecheck(value)
            return super().append(value)

        def insert(self, idx, value):
            self.__typecheck(value)
            return super().insert(idx, value)

        def extend(self, other):
            self.__typecheck(*other)
            return super().extend(other)

    return TypedList


class DirectedGraph:
    def __init__(self):
        self.__edges = {}

    def add_edge(self, node, child):
        self.add_node(node)
        self.add_node(child)
        self.__edges[node].add(child)

    def add_node(self, node):
        if node not in self.__edges:
            self.__edges[node] = set()

    def remove_edge(self, node, child):
        if node in self.__edges:
            self.__edges[node].discard(child)

    def remove_node(self, node):
        self.__edges.pop(node, None)
        for k, v in self.__edges.items():
            v.discard(node)

    def replace_node(self, node, replacement, merge=False):
        if replacement in self.__edges and not merge:
            raise ValueError(
                "Replacement node '{}' already in graph."
                .format(replacement)
            )

        parents = self.parents(node)
        children = self.children(node)
        self.remove_node(node)

        self.add_node(replacement)
        for p in parents:
            self.add_edge(p, replacement)
        for c in children:
            self.add_edge(replacement, c)

    def edges(self):
        return set(
            (n, c)
            for n, cs in self.__edges.items()
            for c in cs
        )

    def nodes(self):
        return set(self.__edges.keys())

    def children(self, *nodes):
        node_children = set()
        for node in nodes:
            node_children.update(self.__edges.get(node, set()))

        return node_children

    def parents(self, *nodes):
        node_parents = set()
        for parent, children in self.__edges.items():
            if children.intersection(nodes):
                node_parents.add(parent)

        return node_parents

    def ancestors(self, *nodes):
        nodes = set(nodes)

        ancestors = self.parents(*nodes)
        tmp = self.parents(*ancestors)
        while tmp - ancestors:
            ancestors.update(tmp)
            tmp = self.parents(*ancestors)

        return ancestors

    def descendants(self, *nodes):
        nodes = set(nodes)

        descendants = self.children(*nodes)
        tmp = self.children(*descendants)
        while tmp - descendants:
            descendants.update(tmp)
            tmp = self.children(*descendants)

        return descendants

    def leaves(self, *nodes):
        nodes = set(nodes)

        leaves = set()
        if len(nodes) == 0:
            leaves.update(k for k, v in self.__edges.items() if not v)
            return leaves

        leaves = self.leaves()
        node_leaves = set()
        while nodes:
            node_leaves.update(nodes.intersection(leaves))
            nodes.difference_update(node_leaves)
            nodes = self.children(*nodes)

        return node_leaves

    def roots(self, *nodes):
        nodes = set(nodes)

        roots = set()
        if len(nodes) == 0:
            roots.update(self.__edges.keys())
            roots.difference_update(*self.__edges.values())
            return roots

        roots = self.roots()
        node_roots = set()
        while nodes:
            node_roots.update(nodes.intersection(roots))
            nodes.difference_update(node_roots)
            nodes = self.parents(*nodes)

        return node_roots



#! /usr/bin/env python3
# SPDX-License-Identifier: GPL-2.0-or-later

# depthcharge-tools depthchargectl program
# Copyright (C) 2021 Alper Nebi Yasak <alpernebiyasak@gmail.com>
# See COPYRIGHT and LICENSE files for full copyright information.

from depthcharge_tools.depthchargectl import depthchargectl

if __name__ == "__main__":
    depthchargectl.main()



#! /usr/bin/env python3
# SPDX-License-Identifier: GPL-2.0-or-later

# depthcharge-tools subprocess utilities
# Copyright (C) 2020-2022 Alper Nebi Yasak <alpernebiyasak@gmail.com>
# See COPYRIGHT and LICENSE files for full copyright information.

import contextlib
import logging
import re
import subprocess
import shlex

from pathlib import Path

logger = logging.getLogger(__name__)


class ProcessRunner:
    def __init__(self, *args_prefix, **kwargs_defaults):
        self.args_prefix = args_prefix
        self.kwargs_defaults = {
            'encoding': "utf-8",
            'check': True,
        }
        self.kwargs_defaults.update(kwargs_defaults)

    def __call__(self, *args_suffix, **kwargs_overrides):
        args = (*self.args_prefix, *args_suffix)
        kwargs = {**self.kwargs_defaults, **kwargs_overrides}

        with contextlib.ExitStack() as ctx:
            stdin = kwargs.get("stdin", None)
            if isinstance(stdin, str):
                stdin = Path(stdin)
            if isinstance(stdin, bytes):
                kwargs["stdin"] = None
                kwargs["encoding"] = None
                kwargs["input"] = stdin
            if isinstance(stdin, Path):
                kwargs["stdin"] = ctx.enter_context(stdin.open("r"))
            if stdin is None:
                kwargs["stdin"] = subprocess.PIPE

            stdout = kwargs.get("stdout", None)
            if isinstance(stdout, str):
                stdout = Path(stdout)
            if isinstance(stdout, Path):
                kwargs["stdout"] = ctx.enter_context(stdout.open("x"))
            if stdout is None:
                kwargs["stdout"] = subprocess.PIPE

            stderr = kwargs.get("stderr", None)
            if isinstance(stderr, str):
                stderr = Path(stderr)
            if isinstance(stderr, Path):
                kwargs["stderr"] = ctx.enter_context(stderr.open("x"))
            if stderr is None:
                kwargs["stderr"] = subprocess.PIPE

            try:
                return subprocess.run(args, **kwargs)
            except subprocess.CalledProcessError as err:
                our_err = self._parse_subprocess_error(err)
                if our_err is None:
                    return subprocess.CompletedProcess(
                        args=err.cmd,
                        returncode=err.returncode,
                        stdout=err.stdout,
                        stderr=err.stderr,
                    )
                if our_err is not err:
                    raise our_err
                raise

    def _parse_subprocess_error(self, err):
        return err


class GzipRunner(ProcessRunner):
    def __init__(self):
        super().__init__("gzip", encoding=None)

    def compress(self, src, dest=None):
        proc = self("-c", "-6", stdin=src, stdout=dest)

        if dest is None:
            return proc.stdout
        else:
            return Path(dest)

    def decompress(self, src, dest=None):
        proc = self("-c", "-d", stdin=src, stdout=dest)

        if dest is None:
            return proc.stdout
        else:
            return Path(dest)

    def test(self, path):
        proc = self("-t", stdin=path, check=False)
        return proc.returncode == 0


class Lz4Runner(ProcessRunner):
    def __init__(self):
        super().__init__("lz4", encoding=None)

    def compress(self, src, dest=None):
        proc = self("-z", "-9", stdin=src, stdout=dest)

        if dest is None:
            return proc.stdout
        else:
            return Path(dest)

    def decompress(self, src, dest=None):
        proc = self("-d", stdin=src, stdout=dest)

        if dest is None:
            return proc.stdout
        else:
            return Path(dest)

    def test(self, path):
        proc = self("-t", stdin=path, check=False)
        return proc.returncode == 0


class LzmaRunner(ProcessRunner):
    def __init__(self):
        super().__init__("lzma", encoding=None)

    def compress(self, src, dest=None):
        proc = self("-z", stdin=src, stdout=dest)

        if dest is None:
            return proc.stdout
        else:
            return Path(dest)

    def decompress(self, src, dest=None):
        proc = self("-d", stdin=src, stdout=dest)

        if dest is None:
            return proc.stdout
        else:
            return Path(dest)

    def test(self, path):
        proc = self("-t", stdin=path, check=False)
        return proc.returncode == 0


class LzopRunner(ProcessRunner):
    def __init__(self):
        super().__init__("lzop", encoding=None)

    def compress(self, src, dest=None):
        proc = self("-c", stdin=src, stdout=dest)

        if dest is None:
            return proc.stdout
        else:
            return Path(dest)

    def decompress(self, src, dest=None):
        proc = self("-c", "-d", stdin=src, stdout=dest)

        if dest is None:
            return proc.stdout
        else:
            return Path(dest)

    def test(self, path):
        proc = self("-t", stdin=path, check=False)
        return proc.returncode == 0


class Bzip2Runner(ProcessRunner):
    def __init__(self):
        super().__init__("bzip2", encoding=None)

    def compress(self, src, dest=None):
        proc = self("-c", stdin=src, stdout=dest)

        if dest is None:
            return proc.stdout
        else:
            return Path(dest)

    def decompress(self, src, dest=None):
        proc = self("-c", "-d", stdin=src, stdout=dest)

        if dest is None:
            return proc.stdout
        else:
            return Path(dest)

    def test(self, path):
        proc = self("-t", stdin=path, check=False)
        return proc.returncode == 0


class XzRunner(ProcessRunner):
    def __init__(self):
        super().__init__("xz", encoding=None)

    def compress(self, src, dest=None):
        proc = self("-z", "--check=crc32", stdin=src, stdout=dest)

        if dest is None:
            return proc.stdout
        else:
            return Path(dest)

    def decompress(self, src, dest=None):
        proc = self("-d", stdin=src, stdout=dest)

        if dest is None:
            return proc.stdout
        else:
            return Path(dest)

    def test(self, path):
        proc = self("-t", stdin=path, check=False)
        return proc.returncode == 0


class ZstdRunner(ProcessRunner):
    def __init__(self):
        super().__init__("zstd", encoding=None)

    def compress(self, src, dest=None):
        proc = self("-z", "-9", stdin=src, stdout=dest)

        if dest is None:
            return proc.stdout
        else:
            return Path(dest)

    def decompress(self, src, dest=None):
        proc = self("-d", stdin=src, stdout=dest)

        if dest is None:
            return proc.stdout
        else:
            return Path(dest)

    def test(self, path):
        proc = self("-t", stdin=path, check=False)
        return proc.returncode == 0


class MkimageRunner(ProcessRunner):
    def __init__(self):
        super().__init__("mkimage")


class VbutilKernelRunner(ProcessRunner):
    def __init__(self):
        super().__init__("futility", "vbutil_kernel")


class CgptRunner(ProcessRunner):
    def __init__(self):
        super().__init__("cgpt")

    def __call__(self, *args, **kwargs):
        proc = super().__call__(*args, **kwargs)
        lines = proc.stdout.splitlines()

        # Sometimes cgpt prints duplicate output.
        # https://bugs.chromium.org/p/chromium/issues/detail?id=463414
        mid = len(lines) // 2
        if lines[:mid] == lines[mid:]:
            proc.stdout = "\n".join(lines[:mid])

        return proc

    def _parse_subprocess_error(self, err):
        # Exits with nonzero status if it finds no partitions of
        # given type even if the disk has a valid partition table
        if not err.stderr:
            return None

        m = re.fullmatch(
            "ERROR: Can't open (.*): Permission denied\n",
            err.stderr,
        )
        if m:
            return PermissionError(
                "Couldn't open '{}', permission denied."
                .format(m.groups()[0])
            )

        return err


    def get_raw_attribute(self, disk, partno):
        proc = self("show", "-A", "-i", str(partno), str(disk))
        attribute = int(proc.stdout, 16)
        return attribute

    def set_raw_attribute(self, disk, partno, attribute):
        self("add", "-A", hex(attribute), "-i", str(partno), str(disk))

    def get_flags(self, disk, partno):
        attribute = self.get_raw_attribute(disk, partno)
        successful = (attribute >> 8) & 0x1
        tries = (attribute >> 4) & 0xF
        priority = (attribute >> 0) & 0xF

        return {
            "A": attribute,
            "S": successful,
            "P": priority,
            "T": tries,
        }

    def set_flags(self, disk, partno, A=None, S=None, P=None, T=None):
        flag_args = []
        if A is not None:
            flag_args += ["-A", str(int(A))]
        if S is not None:
            flag_args += ["-S", str(int(S))]
        if P is not None:
            flag_args += ["-P", str(int(P))]
        if T is not None:
            flag_args += ["-T", str(int(T))]

        self("add", *flag_args, "-i", str(partno), str(disk))

    def get_size(self, disk, partno):
        proc = self("show", "-s", "-i", str(partno), str(disk))
        blocks = int(proc.stdout)
        return blocks * 512

    def get_start(self, disk, partno):
        proc = self("show", "-b", "-i", str(partno), str(disk))
        blocks = int(proc.stdout)
        return blocks * 512

    def find_partitions(self, disk, type=None):
        if type is None:
            # cgpt find needs at least one of -t, -u, -l
            proc = self("show", "-q", "-n", disk)
            lines = proc.stdout.splitlines()
            partnos = [int(shlex.split(line)[2]) for line in lines]

        else:
            proc = self("find", "-n", "-t", type, disk)
            partnos = [int(n) for n in proc.stdout.splitlines()]

        return partnos

    def prioritize(self, disk, partno):
        self("prioritize", "-i", str(partno), str(disk))


class CrossystemRunner(ProcessRunner):
    def __init__(self):
        super().__init__("crossystem")

    def hwid(self):
        proc = self("hwid", check=False)

        if proc.returncode == 0:
            return proc.stdout
        else:
            return None

    def fwid(self):
        proc = self("fwid", check=False)

        if proc.returncode == 0:
            return proc.stdout
        else:
            return None


class FdtgetRunner(ProcessRunner):
    def __init__(self):
        super().__init__("fdtget")

    def get(self, dt_file, node='/', prop='', default=None, type=None):
        options = []

        if default is not None:
            options += ["--default", str(default)]

        if type == str:
            options += ["--type", "s"]
        elif type == int:
            options += ["--type", "i"]
        elif type == bytes:
            options += ["--type", 'bx']
        elif type is not None:
            options += ["--type", str(type)]

        proc = self(*options, str(dt_file), str(node), str(prop))

        # str.split takes too much memory
        def split(s):
            for m in re.finditer("(\S*)\s", s):
                yield m.group()

        if type in (None, int):
            try:
                data = [int(i) for i in split(proc.stdout)]
                return data[0] if len(data) == 1 else data
            except:
                pass

        if type in (None, bytes):
            try:
                # bytes.fromhex("0") doesn't work
                data = bytes(int(x, 16) for x in split(proc.stdout))
                return data
            except:
                pass

        data = str(proc.stdout).strip("\n")
        return data

    def properties(self, dt_file, node='/'):
        proc = self("--properties", str(dt_file), str(node), check=False)

        if proc.returncode == 0:
            return proc.stdout.splitlines()
        else:
            return []

    def subnodes(self, dt_file, node='/'):
        proc = self("--list", str(dt_file), str(node), check=False)
        nodes = proc.stdout.splitlines()

        if proc.returncode == 0:
            return proc.stdout.splitlines()
        else:
            return []


class FdtputRunner(ProcessRunner):
    def __init__(self):
        super().__init__("fdtput")

    def put(self, dt_file, node='/', prop='', value=None, type=None):
        if isinstance(value, list):
            values = value
        else:
            values = [value]

        value_args = []
        for value in values:
            if isinstance(value, str):
                value_args.append(value)
                if type is None:
                    type = str

            elif isinstance(value, bytes):
                value_args.extend(hex(c) for c in value)
                if type is None:
                    type = bytes

            elif isinstance(value, int):
                value_args.append(str(value))
                if type is None:
                    type = int

            else:
                value_args.append(str(value))

        options = []
        if type == str:
            options += ["--type", "s"]
        elif type == int:
            options += ["--type", "i"]
        elif type == bytes:
            options += ["--type", 'bx']
        elif type is not None:
            options += ["--type", str(type)]

        self(*options, str(dt_file), str(node), str(prop), *value_args)


class FileRunner(ProcessRunner):
    def __init__(self):
        super().__init__("file")

    def brief(self, path):
        proc = self("-b", path, check=False)

        if proc.returncode == 0:
            return proc.stdout.strip("\n")
        else:
            return None


gzip = GzipRunner()
lz4 = Lz4Runner()
lzma = LzmaRunner()
lzop = LzopRunner()
bzip2 = Bzip2Runner()
xz = XzRunner()
zstd = ZstdRunner()
mkimage = MkimageRunner()
vbutil_kernel = VbutilKernelRunner()
cgpt = CgptRunner()
crossystem = CrossystemRunner()
fdtget = FdtgetRunner()
fdtput = FdtputRunner()
file = FileRunner()



#! /usr/bin/env python3
# SPDX-License-Identifier: GPL-2.0-or-later

# depthcharge-tools __init__.py
# Copyright (C) 2020-2021 Alper Nebi Yasak <alpernebiyasak@gmail.com>
# See COPYRIGHT and LICENSE files for full copyright information.

import glob
import logging
import pathlib
import pkg_resources
import re
import subprocess

logger = logging.getLogger(__name__)
logger.addHandler(logging.NullHandler())


def get_version():
    version = None
    pkg_path = pkg_resources.resource_filename(__name__, '')
    pkg_path = pathlib.Path(pkg_path).resolve()

    try:
        self = pkg_resources.require(__name__)[0]
        version = self.version

    except pkg_resources.DistributionNotFound:
        setup_py = pkg_path.parent / "setup.py"
        if setup_py.exists():
            version = re.findall(
                'version=(\'.+\'|".+"),',
                setup_py.read_text(),
            )[0].strip('"\'')

    if (pkg_path.parent / ".git").exists():
        proc = subprocess.run(
            ["git", "-C", pkg_path, "describe"],
            stdout=subprocess.PIPE,
            encoding="utf-8",
            check=False,
        )
        if proc.returncode == 0:
            tag, *local = proc.stdout.split("-")

            if local:
                version = "{}+{}".format(tag, ".".join(local))
            else:
                version = tag

    if version is not None:
        return pkg_resources.parse_version(version)

__version__ = get_version()

config_ini = pkg_resources.resource_string(__name__, "config.ini")
config_ini = config_ini.decode("utf-8")

boards_ini = pkg_resources.resource_string(__name__, "boards.ini")
boards_ini = boards_ini.decode("utf-8")

config_files = [
    *glob.glob("/etc/depthcharge-tools/config"),
    *glob.glob("/etc/depthcharge-tools/config.d/*"),
]





#! /usr/bin/env python3
# SPDX-License-Identifier: GPL-2.0-or-later

# depthcharge-tools os utilities
# Copyright (C) 2020-2022 Alper Nebi Yasak <alpernebiyasak@gmail.com>
# See COPYRIGHT and LICENSE files for full copyright information.

import collections
import re
import shlex

from pathlib import Path

from depthcharge_tools.utils.collections import (
    DirectedGraph,
)
from depthcharge_tools.utils.pathlib import (
    iterdir,
    read_lines,
)
from depthcharge_tools.utils.platform import (
    proc_cmdline,
)
from depthcharge_tools.utils.subprocess import (
    cgpt,
)


class Disks(DirectedGraph):
    def __init__(
        self,
        sys="/sys",
        dev="/dev",
        fstab="/etc/fstab",
        mtab="/etc/mtab",
        procmounts="/proc/self/mounts",
        mountinfo="/proc/self/mountinfo",
        crypttab="/etc/crypttab",
    ):
        super().__init__()

        self._sys = sys = Path(sys)
        self._dev = dev = Path(dev)
        self._fstab = fstab = Path(fstab)
        self._procmounts = procmounts = Path(procmounts)
        self._mtab = mtab = Path(mtab)
        self._mountinfo = mountinfo = Path(mountinfo)
        self._crypttab = crypttab = Path(crypttab)

        for sysdir in iterdir(sys / "class" / "block"):
            for device in read_lines(sysdir / "dm" / "name"):
                self.add_edge(dev / sysdir.name, dev / "mapper" / device)

            for device in iterdir(sysdir / "slaves"):
                self.add_edge(dev / device.name, dev / sysdir.name)

            for device in iterdir(sysdir / "holders"):
                self.add_edge(dev / sysdir.name, dev / device.name)

            for device in iterdir(sysdir):
                if device.name.startswith(sysdir.name):
                    self.add_edge(dev / sysdir.name, dev / device.name)

        for line in read_lines(crypttab):
            if line and not line.startswith("#"):
                fields = shlex.split(line)
                cryptdev, device = fields[0], fields[1]
                if device != 'none':
                    cryptdev = dev / "mapper" / cryptdev
                    self.add_edge(device, cryptdev)

        fstab_mounts = {}
        for line in read_lines(fstab):
            if line and not line.startswith("#"):
                fields = shlex.split(line)
                device, mount = fields[0], fields[1]
                if mount != 'none':
                    fstab_mounts[mount] = device

        procmounts_mounts = {}
        for line in read_lines(procmounts):
            if line and not line.startswith("#"):
                fields = shlex.split(line)
                device, mount = fields[0], fields[1]
                device = self.evaluate(device)
                if device is not None:
                    procmounts_mounts[mount] = device

        mtab_mounts = {}
        for line in read_lines(mtab):
            if line and not line.startswith("#"):
                fields = shlex.split(line)
                device, mount = fields[0], fields[1]
                device = self.evaluate(device)
                if device is not None:
                    mtab_mounts[mount] = device

        mountinfo_mounts = {}
        for line in read_lines(mountinfo):
            if line and not line.startswith("#"):
                fields = shlex.split(line)
                device, fsroot, mount = fields[2], fields[3], fields[4]
                if fsroot != "/":
                    mountinfo_mounts[mount] = None
                    continue
                device = self.evaluate(device)
                if device is not None:
                    mountinfo_mounts[mount] = device

        mounts = collections.ChainMap(
            fstab_mounts,
            mountinfo_mounts,
            procmounts_mounts,
            mtab_mounts,
        )

        self._fstab_mounts = fstab_mounts
        self._procmounts_mounts = procmounts_mounts
        self._mtab_mounts = mtab_mounts
        self._mountinfo_mounts = mountinfo_mounts
        self._mounts = mounts

    def __getitem__(self, key):
        return self.evaluate(key)

    def evaluate(self, device):
        dev = self._dev
        sys = self._sys

        if device is None:
            return None

        elif isinstance(device, Path):
            device = str(device)

        elif isinstance(device, (Disk, Partition)):
            device = str(device.path)

        if device.startswith("ID="):
            id_ = device[len("ID="):]
            if not id_:
                return None

            device = dev / "disk" / "by-id" / id_

        elif device.startswith("LABEL="):
            label = device[len("LABEL="):]
            if not label:
                return None

            device = dev / "disk" / "by-label" / label

        elif device.startswith("PARTLABEL="):
            partlabel = device[len("PARTLABEL="):]
            if not partlabel:
                return None

            device = dev / "disk" / "by-partlabel" / partlabel

        elif device.startswith("UUID="):
            uuid = device[len("UUID="):]
            if not uuid:
                return None

            device = dev / "disk" / "by-uuid" / uuid
            if not device.exists():
                device = dev / "disk" / "by-uuid" / uuid.lower()

        elif device.startswith("PARTUUID="):
            partuuid, _, partnroff = (
                device[len("PARTUUID="):].partition("/PARTNROFF=")
            )
            if not partuuid:
                return None

            device = dev / "disk" / "by-partuuid" / partuuid
            if not device.exists():
                device = dev / "disk" / "by-partuuid" / partuuid.lower()

            if partnroff:
                device = device.resolve()
                match = re.match("(.*[^0-9])([0-9]+)$", device.name)
                if not match:
                    return None
                prefix, partno = match.groups()
                partno = str(int(partno) + int(partnroff))
                device = device.with_name("{}{}".format(prefix, partno))

        elif re.match("[0-9]+:[0-9]+", device):
            device = dev / "block" / device

        # Encrypted devices may currently be set up with names different
        # than in the crypttab file, so check that as well.
        elif device.startswith(str(dev / "mapper")):
            if not Path(device).resolve().exists():
                for line in read_lines(self._crypttab):
                    if not line or line.startswith("#"):
                        continue

                    fields = shlex.split(line)
                    parentdev, cryptdev = fields[1], fields[0]
                    if cryptdev != device.split("/")[-1]:
                        continue

                    parentdev = self.evaluate(parentdev)
                    siblings = self.children(parentdev)
                    if len(siblings) == 1:
                        device = str(siblings.pop())

                    # This is actually wrong, but we can't really decide
                    # which to use. The parent's good enough for us since
                    # we usually only care about going up the tree.
                    else:
                        device = str(parentdev)

        device = Path(device).resolve()
        if not device.exists() or dev not in device.parents:
            return None

        try:
            return Partition(device, dev=dev, sys=sys)
        except:
            pass

        try:
            return Disk(device, dev=dev, sys=sys)
        except:
            pass

    def by_mountpoint(self, mountpoint, fstab_only=False):
        if not Path(mountpoint).exists():
            return None

        if fstab_only:
            # We want the form in the fstab, e.g. PARTUUID=*
            device = self._fstab_mounts.get(mountpoint)
            return device
        else:
            device = self._mounts.get(str(mountpoint))
            return self.evaluate(device)

    def mountpoints(self, device, include_fstab=False):
        device = self.evaluate(device)
        if device is None:
            return set()

        # Exclude fstab whose entries are not necessarily mounted
        if not include_fstab:
            mounts = collections.ChainMap(
                self._mountinfo_mounts,
                self._procmounts_mounts,
                self._mtab_mounts,
            )
        else:
            mounts = self._mounts

        mountpoints = set()
        for mnt, dev in mounts.items():
            dev = self.evaluate(dev)
            if dev == device:
                mnt = Path(mnt).resolve()
                if mnt.exists() or include_fstab:
                    mountpoints.add(mnt)

        return mountpoints

    def by_id(self, id_):
        return self.evaluate("ID={}".format(id_))

    def by_label(self, label):
        return self.evaluate("LABEL={}".format(label))

    def by_partlabel(self, partlabel):
        return self.evaluate("PARTLABEL={}".format(partlabel))

    def by_uuid(self, uuid):
        return self.evaluate("UUID={}".format(uuid))

    def by_partuuid(self, partuuid):
        return self.evaluate("PARTUUID={}".format(partuuid))

    def _get_dev_disk_info(self, device, prop):
        device = self.evaluate(device)
        for path in self._dev.glob("disk/by-{}/*".format(prop)):
            dev = self.evaluate(path)
            if dev == device:
                return path.name

    def get_id(self, device):
        return self._get_dev_disk_info(device, "id")

    def get_label(self, device):
        return self._get_dev_disk_info(device, "label")

    def get_partlabel(self, device):
        return self._get_dev_disk_info(device, "partlabel")

    def get_uuid(self, device):
        return self._get_dev_disk_info(device, "uuid")

    def get_partuuid(self, device):
        return self._get_dev_disk_info(device, "partuuid")

    def by_kern_guid(self):
        for arg in proc_cmdline():
            lhs, _, rhs = arg.partition("=")
            if lhs == "kern_guid":
                return self.by_partuuid(rhs)

    def add_edge(self, node, child):
        node = self.evaluate(node)
        child = self.evaluate(child)
        if node is not None and child is not None and node != child:
            return super().add_edge(node, child)

    def children(self, *nodes):
        return super().children(*map(self.evaluate, nodes))

    def parents(self, *nodes):
        return super().parents(*map(self.evaluate, nodes))

    def leaves(self, *nodes):
        return super().leaves(*map(self.evaluate, nodes))

    def roots(self, *nodes):
        return super().roots(*map(self.evaluate, nodes))


class Disk:
    def __init__(self, path, dev="/dev", sys="/sys"):
        self._sys = sys = Path(sys)
        self._dev = dev = Path(dev)

        if isinstance(path, Disk):
            path = path.path
        else:
            path = Path(path).resolve()

        if not (path.is_file() or path.is_block_device()):
            fmt = "Disk '{}' is not a file or block device."
            msg = fmt.format(str(path))
            raise ValueError(msg)

        self.path = path

    def partition(self, partno):
        return Partition(self, partno, dev=self._dev, sys=self._sys)

    def partitions(self):
        return [
            Partition(self, n, dev=self._dev, sys=self._sys)
            for n in cgpt.find_partitions(self.path)
        ]

    def cros_partitions(self):
        return [
            CrosPartition(self, n, dev=self._dev, sys=self._sys)
            for n in cgpt.find_partitions(self.path, type="kernel")
        ]

    @property
    def size(self):
        if self.path.is_file():
            return self.path.stat().st_size

        if self.path.is_block_device():
            sysdir = self._sys / "class" / "block" / self.path.name

            size_f = sysdir / "size"
            if size_f.exists():
                blocks = int(size_f.read_text())
                return blocks * 512

    def __hash__(self):
        return hash((self.path,))

    def __eq__(self, other):
        if isinstance(other, Disk):
            return self.path == other.path
        return False

    def __str__(self):
        return str(self.path)

    def __repr__(self):
        cls = self.__class__.__name__
        return "{}('{}')".format(cls, self.path)


class Partition:
    def __init__(self, path, partno=None, dev="/dev", sys="/sys"):
        self._dev = dev = Path(dev)
        self._sys = sys = Path(sys)

        if isinstance(path, Disk):
            disk = path
            path = None
        elif isinstance(path, Partition):
            disk = path.disk
            partno = path.partno
            path = path.path
        else:
            disk = None
            path = Path(path).resolve()

        if (
            disk is None
            and partno is None
            and path.parent == dev
            and path.is_block_device()
        ):
            match = (
                re.fullmatch("(.*[0-9])p([0-9]+)", path.name)
                or re.fullmatch("(.*[^0-9])([0-9]+)", path.name)
            )
            if match:
                diskname, partno = match.groups()
                partno = int(partno)
                disk = Disk(path.with_name(diskname), dev=dev, sys=sys)

        if disk is None:
            disk = Disk(path, dev=dev, sys=sys)
            path = None

        if partno is None:
            fmt = "Partition number not given for disk '{}'."
            msg = fmt.format(str(disk))
            raise ValueError(msg)

        elif not (isinstance(partno, int) and partno > 0):
            fmt = "Partition number '{}' must be a positive integer."
            msg = fmt.format(partno)
            raise ValueError(msg)

        elif (
            path is None
            and disk.path.parent == dev
            and disk.path.is_block_device()
        ):
            fmt = "{}p{}" if disk.path.name[-1].isnumeric() else "{}{}"
            name = fmt.format(disk.path.name, partno)
            path = disk.path.with_name(name)

        if path is not None:
            if not (path.is_file() or path.is_block_device()):
                path = None

        self.disk = disk
        self.path = path
        self.partno = partno

    @property
    def size(self):
        if self.path is None:
            return cgpt.get_size(self.disk.path, self.partno)

        if self.path.is_file():
            return self.path.stat().st_size

        if self.path.is_block_device():
            sysdir = self._sys / "class" / "block" / self.path.name

            size_f = sysdir / "size"
            if size_f.exists():
                blocks = int(size_f.read_text())
                return blocks * 512

    def write_bytes(self, data):
        data = bytes(data)

        if len(data) >= self.size:
            raise ValueError(
                "Data to be written ('{}' bytes) is bigger than "
                "partition '{}' ('{}' bytes)."
                .format(len(data), self, self.size)
            )

        if self.path is None:
            start = cgpt.get_start(self.disk.path, self.partno)

            with self.disk.path.open("r+b") as disk:
                seek = disk.seek(start)
                if seek != start:
                    raise IOError(
                        "Couldn't seek disk to start of partition '{}'."
                        .format(self)
                    )

                written = disk.write(data)
                if written != len(data):
                    raise IOError(
                        "Couldn't write data to partition '{}' "
                        "(wrote '{}' out of '{}' bytes)."
                        .format(self, written, len(data))
                    )

        else:
            self.path.write_bytes(data)

    def __hash__(self):
        return hash((self.path, self.disk, self.partno))

    def __eq__(self, other):
        if isinstance(other, Partition):
            return (
                self.path == other.path
                and self.disk == other.disk
                and self.partno == other.partno
            )
        return False

    def __str__(self):
        if self.path is not None:
            return str(self.path)
        else:
            return "{}#{}".format(self.disk.path, self.partno)

    def __repr__(self):
        cls = self.__class__.__name__
        if self.path is not None:
            return "{}('{}')".format(cls, self.path)
        else:
            return "{}('{}', {})".format(cls, self.disk.path, self.partno)


class CrosPartition(Partition):
    @property
    def attribute(self):
        return cgpt.get_raw_attribute(self.disk.path, self.partno)

    @attribute.setter
    def attribute(self, attr):
        return cgpt.set_raw_attribute(self.disk.path, self.partno, attr)

    @property
    def flags(self):
        flags = cgpt.get_flags(self.disk.path, self.partno)
        return {
            "attribute": flags["A"],
            "successful": flags["S"],
            "priority": flags["P"],
            "tries": flags["T"],
        }

    @flags.setter
    def flags(self, value):
        if isinstance(value, dict):
            A = value.get("attribute", None)
            S = value.get("successful", None)
            P = value.get("priority", None)
            T = value.get("tries", None)

        else:
            A = getattr(value, "attribute", None)
            S = getattr(value, "successful", None)
            P = getattr(value, "priority", None)
            T = getattr(value, "tries", None)

        cgpt.set_flags(self.disk.path, self.partno, A=A, S=S, P=P, T=T)

    @property
    def successful(self):
        return self.flags["successful"]

    @successful.setter
    def successful(self, value):
        self.flags = {"successful": value}

    @property
    def tries(self):
        return self.flags["tries"]

    @tries.setter
    def tries(self, value):
        self.flags = {"tries": value}

    @property
    def priority(self):
        return self.flags["priority"]

    @priority.setter
    def priority(self, value):
        self.flags = {"priority": value}

    def prioritize(self):
        return cgpt.prioritize(self.disk.path, self.partno)

    def _comparable_parts(self):
        flags = self.flags
        size = self.size

        return (
            flags["successful"],
            flags["priority"],
            flags["tries"],
            self.size,
        )

    def __lt__(self, other):
        if not isinstance(other, CrosPartition):
            return NotImplemented

        return self._comparable_parts() < other._comparable_parts()

    def __gt__(self, other):
        if not isinstance(other, CrosPartition):
            return NotImplemented

        return self._comparable_parts() > other._comparable_parts()



# SPDX-License-Identifier: GPL-2.0-or-later

# depthcharge-tools boards.ini update script
# Copyright (C) 2021-2022 Alper Nebi Yasak <alpernebiyasak@gmail.com>
# See COPYRIGHT and LICENSE files for full copyright information.

#! /usr/bin/env python3

import configparser
import collections
import json
import logging
import re

from functools import lru_cache
from pathlib import Path

from depthcharge_tools import __version__
from depthcharge_tools.utils.argparse import (
    Command,
    Argument,
    Group,
)
from depthcharge_tools.utils.collections import (
    DirectedGraph,
    SortedDict,
)
from depthcharge_tools.utils.pathlib import (
    iterdir,
    read_lines,
)


class update_config(
    Command,
    prog="update_config.py",
    add_help=False,
):
    """
    Maintainer tool to help update depthcharge-tools config.ini

    ---

    If you're packaging depthcharge-tools, don't use this as a build
    step. Results from this are intended to be checked and modified
    manually before they go into the final config.ini, the file
    committed to the repository is the canonical one.
    """

    logger = logging.getLogger(__name__)

    @Group
    def options(self):
        """Options"""

    @options.add
    @Argument("-h", "--help", action="help")
    def print_help(self):
        """Show this help message."""
        # type(self).parser.print_help()

    @options.add
    @Argument(
        "-V", "--version",
        action="version",
        version="depthcharge-tools %(prog)s {}".format(__version__),
    )
    def version(self):
        """Print program version."""
        return type(self).version.version % {"prog": type(self).prog}

    @options.add
    @Argument("-v", "--verbose", count=True)
    def verbosity(self, verbosity=0):
        """Print more detailed output."""
        level = logging.WARNING - int(verbosity) * 10
        self.logger.setLevel(level)
        return verbosity

    def parse_recovery_conf_block(self, block):
        values = {}

        for line in block.splitlines():
            if line.startswith("#"):
                continue

            key, sep, value = line.partition("=")
            if sep != "=":
                raise ValueError(
                    "No equals sign in line: '{}'"
                    .format(line)
                )

            if key not in values:
                values[key] = value
            elif isinstance(values[key], list):
                values[key].append(value)
            else:
                values[key] = [values[key], value]

        if "filesize" in values:
            values["filesize"] = int(values["filesize"] or 0)
        if "zipfilesize" in values:
            values["zipfilesize"] = int(values["zipfilesize"] or 0)

        return values

    @options.add
    @Argument("-r", "--recovery-conf")
    def recovery_conf(self, path=None):
        """\
        Chrome OS recovery.conf file for their Linux recovery tool

        https://dl.google.com/dl/edgedl/chromeos/recovery/recovery.conf
        """
        return Path(path) if path else None

    @property
    @lru_cache
    def recovery_conf_boards(self):
        if self.recovery_conf is None:
            return {}

        header, *blocks = [
            self.parse_recovery_conf_block(block)
            for block in re.split("\n\n+", self.recovery_conf.read_text())
        ]

        version = header.get(
            "recovery_tool_linux_version",
            header.get("recovery_tool_version"),
        )

        if version != "0.9.2":
            raise TypeError(
                "Unsupported recovery.conf version: {}"
                .format(header.get("recovery_tool_update", version))
            )

        boards = collections.defaultdict(list)
        for block in blocks:
            hwidmatch = block.get("hwidmatch")

            # This might be a parent board, but the best fallback we have
            codename = block.get("file").split("_")[2]

            if hwidmatch == "duplicate of rabbid":
                codename = "rabbid"
                block["hwidmatch"] = None
            elif hwidmatch == "duplicate of C433":
                codename = "shyvana"
                block["hwidmatch"] = None
            elif hwidmatch == "Duplicate of BARLA":
                codename = "barla"
                block["hwidmatch"] = None

            elif hwidmatch.strip("^(").startswith("ACER ZGB"):
                pass # x86-zgb, x86-zgb-he
            elif hwidmatch.strip("^(").startswith("IEC MARIO"):
                pass # x86-mario
            elif hwidmatch.strip("^(").startswith("SAMS ALEX"):
                pass # x86-alex, x86-alex-he

            elif hwidmatch in (
                "DOES NOT MATCH ANYTHING",
                "NO MATCH JUST FOR ENTRY",
            ):
                codename = block.get("file").split("_")[2]
                block["hwidmatch"] = None

            else:
                m = re.match("^\^?\(?([0-9A-Z]+)[^0-9A-Za-z]", hwidmatch)
                if m:
                    codename = m.group(1).lower()
                else:
                    self.logger.warning(
                        "Could not parse codename for hwidmatch '{}'."
                        .format(hwidmatch)
                    )

            if codename:
                boards[codename].append(block)

        return dict(boards)

    def read_profiles_repo_name(self, d):
        # A single-line file, so return the first line
        for line in read_lines(d / "profiles" / "repo_name"):
            return line.strip()

    def parse_layout_conf(self, d):
        values = {}

        for line in read_lines(d / "metadata" / "layout.conf"):
            key, eq, value = line.partition("=")
            if eq == "=" and "#" not in key:
                values[key.strip()] = value.strip()

        return values

    def get_profiles_base_parent_boards(self, d):
        parents = []

        for line in read_lines(d / "profiles" / "base" / "parent"):
            # Most end with :base, but there were e.g. freon:base/amd64
            lhs, sep, rhs = line.partition(":")
            if sep == ":" and rhs.startswith("base"):
                parents.append(lhs)

            # Very old scheme, e.g. firmware-snow-2695.B tegra variants
            prefix = "../../../"
            suffix = "/profiles/base"
            if line.startswith(prefix) and line.endswith(suffix):
                parents.append(line[len(prefix):-len(suffix)])

        return parents

    def get_model_yaml_boards(self, d):
        children = set()

        # chromeos-config-bsp directories can have inconsistent names.
        for config_d in d.glob("chromeos-base/chromeos-config-bsp*"):
            for line in read_lines(config_d / "files" / "model.yaml"):
                # A giant hack that lets me avoid parsing yaml
                keyname = "- $device-name:"
                space, sep, child = line.partition(keyname)
                if sep == keyname and "#" not in space:
                    children.add(child.strip().strip('\'"'))

        return children

    @options.add
    @Argument("-b", "--board-overlays-repo")
    def board_overlays_repo(self, path=None):
        """\
        Chromium OS board-overlays git repository

        https://chromium.googlesource.com/chromiumos/overlays/board-overlays
        """
        return Path(path) if path else None

    def get_project_config_boards(self, d):
        children = set()

        project_config = (
            d / "sw_build_config" / "platform" / "chromeos-config"
            / "generated" / "project-config.json"
        )
        if project_config.is_file():
            config = json.loads(project_config.read_text())

            for section in config["chromeos"]["configs"]:
                if section["name"]:
                    children.add(section["name"])

        return children

    @options.add
    @Argument("-p", "--chromiumos-project-repo")
    def chromiumos_project_repo(self, path=None):
        """\
        Chromium OS's chromiumos/project git repository

        https://chromium.googlesource.com/chromiumos/project
        """
        return Path(path) if path else None

    def parse_defconfig(self, text):
        values = dict()

        for line in text.splitlines():
            if line.startswith("#"):
                continue

            lhs, sep, rhs = line.partition("=")
            if sep != "=" or not lhs.startswith("CONFIG_"):
                continue

            if rhs == "y":
                value = True
            elif rhs == "n":
                value = False
            elif rhs.startswith("0x"):
                value = int(rhs, 16)
            else:
                value = rhs.strip().strip("'\"")

            key = lhs[len("CONFIG_"):]
            values[key] = value

        # Not going to parse Kconfig for this
        if values.get("ARCH_ARM") and not values.get("ARCH_ARM_V8"):
            values["ARCH_ARM_V7"] = True

        return values

    @options.add
    @Argument("-d", "--depthcharge-repo")
    def depthcharge_repo(self, path=None):
        """\
        Chromium OS depthcharge firmware git repository

        https://chromium.googlesource.com/chromiumos/platform/depthcharge
        """
        return Path(path) if path else None

    def parse_kconfig_defaults(self, text):
        defaults = {}

        clean_text, _ = re.subn("#.*\n", "\n", text)
        blocks = re.split("\n\n+", clean_text)
        for block in blocks:
            config = None

            for line in block.splitlines():
                line = line.strip()

                if not line or line.startswith("help"):
                    if config is None:
                        continue
                    else:
                        config = None
                        break

                m = re.match("config ([0-9A-Z_]+)", line)
                if m:
                    config = m.group(1)
                    type_ = lambda s: str.strip(s, "'\"")
                    defaults[config] = {}

                if config is None:
                    continue

                if line.startswith("hex"):
                    type_ = lambda x: int(x, 16)
                elif line.startswith("int"):
                    type_ = int
                elif line.startswith("bool"):
                    type_ = lambda b: b in ("y", "Y")
                elif line.startswith("string"):
                    type_ = lambda s: str.strip(s, "'\"")

                m = re.match("default (\S+|\".+\")$", line)
                try:
                    value = type_(m.group(1).strip("'\""))
                except ValueError:
                    value = m.group(1)
                except AttributeError:
                    value = None
                finally:
                    if value is not None:
                        defaults[config][None] = value
                    value = None

                m = re.match("default (\S+|\".+\") if ([0-9A-Z_]+)", line)
                try:
                    value = type_(m.group(1))
                    cond = m.group(2)
                except ValueError:
                    value = m.group(1)
                    cond = m.group(2)
                except AttributeError:
                    value = None
                    cond = None
                finally:
                    if value is not None and cond is not None:
                        defaults[config][cond] = value
                    value = None

        return defaults

    def parse_kconfig_selects(self, text):
        selects = {}

        clean_text, _ = re.subn("#.*\n", "\n", text)
        blocks = re.split("\n\n+", clean_text)
        for block in blocks:
            config = None

            for line in block.splitlines():
                line = line.strip()

                if not line or line.startswith("help"):
                    if config is None:
                        continue
                    else:
                        config = None
                        break

                m = re.match("config ([0-9A-Z_]+)", line)
                if m:
                    config = m.group(1)
                    type_ = lambda s: str.strip(s, "'\"")
                    selects[config] = {}
                    selects[config][None] = []

                if config is None:
                    continue

                m = re.match("select (\S+|\".+\")$", line)
                if m:
                    value = m.group(1).strip("'\"")
                    selects[config][None].append(value)

                m = re.match("select (\S+|\".+\") if ([0-9A-Z_]+)", line)
                if m:
                    value = m.group(1)
                    cond = m.group(2)
                    if cond not in selects[config]:
                        selects[config][cond] = []
                    selects[config][cond].append(value)

        return selects

    @property
    @lru_cache
    def depthcharge_boards(self):
        boards = {}
        defaults = collections.defaultdict(dict)

        if self.depthcharge_repo is None:
            return boards

        # Provide a limited set of default values to avoid having to
        # parse all Kconfig files or something
        image_f = self.depthcharge_repo / "src/image/Kconfig"
        image_d = self.parse_kconfig_defaults(image_f.read_text())

        for cond, default in image_d.get("KERNEL_SIZE", {}).items():
            defaults[cond]["KERNEL_SIZE"] = default

        for defconfig_f in self.depthcharge_repo.glob("board/*/defconfig"):
            defconfig = self.parse_defconfig(defconfig_f.read_text())

            # CONFIG_BOARD is removed in master
            board = defconfig.get("BOARD", defconfig_f.parent.name)

            # kevin, kevin-tpm2 both have BOARD="kevin", prefer former
            if board in boards and board != defconfig_f.parent.name:
                continue

            board_d = {}
            board_d.update(defaults.get(None, {}))
            for cond, config in defaults.items():
                if cond and defconfig.get(cond, None):
                    board_d.update(config)

            board_d.update(defconfig)
            boards[board] = board_d

        return boards

    @options.add
    @Argument("-c", "--coreboot-repo")
    def coreboot_repo(self, path=None):
        """\
        Chromium OS coreboot firmware git repository

        https://chromium.googlesource.com/chromiumos/third_party/coreboot
        """
        return Path(path) if path else None

    @property
    @lru_cache
    def coreboot_boards(self):
        boards = {}

        if self.coreboot_repo is None:
            return boards

        def get_board_name(config):
            parts = config.split("_")
            if len(parts) < 2 or parts[0] != "BOARD":
                return None

            vendor = parts[1].lower()
            if not (self.coreboot_repo / "src/mainboard" / vendor).is_dir():
                return None

            board = "_".join(config.split("_")[2:]).lower()
            return board

        for kconfig_f in self.coreboot_repo.glob("src/mainboard/*/*/Kconfig"):
            kconfig_name = kconfig_f.with_name("Kconfig.name")
            kconfig = kconfig_f.read_text()
            if kconfig_name.is_file():
                kconfig_name = kconfig_name.read_text()
            else:
                kconfig_name = ""

            defaults = self.parse_kconfig_defaults(kconfig)
            selects = self.parse_kconfig_selects(kconfig)
            selects.update(self.parse_kconfig_selects(kconfig_name))

            def add_board(config):
                board = get_board_name(config)
                if board in boards:
                    return boards[board]

                boards[board] = {}

                for cond, selectlist in selects.get(config, {}).items():
                    if cond is None or cond in boards[board]:
                        for select in selectlist:
                            if get_board_name(select):
                                boards[board].update(add_board(select))
                            boards[board][select] = True

                for key, values in defaults.items():
                    if get_board_name(key):
                        continue
                    value = values.get(config, values.get(None))
                    if value is not None:
                        boards[board][key] = value

                board_opts = selects.get("BOARD_SPECIFIC_OPTIONS", {})
                for cond, selectlist in board_opts.items():
                    if cond is None or cond in boards[board]:
                        for select in selectlist:
                            if get_board_name(select):
                                boards[board].update(add_board(select))
                            boards[board][select] = True

                boards[board][config] = True

                return boards[board]

            for config, _ in defaults.items():
                if get_board_name(config):
                    add_board(config)

            for select, _ in selects.items():
                if get_board_name(select):
                    add_board(select)

        for board, block in list(boards.items()):
            suffix = "_common"
            if board.endswith(suffix):
                actual = board[:-len(suffix)]
                boards.setdefault(actual, boards.pop(board))
                board = actual

            prefix = "baseboard_"
            if board.startswith(prefix):
                actual = "baseboard-{}".format(board[len(prefix):])
                boards.setdefault(actual, boards.pop(board))
                board = actual

            if not block.get("MAINBOARD_HAS_CHROMEOS", False):
                if board in boards:
                    boards.pop(board)
                continue

        return boards

    @property
    @lru_cache
    def board_relations(self):
        board_relations = DirectedGraph()
        repo_names = {}

        # Find canonical names for each board
        for board_d in iterdir(self.board_overlays_repo):
            if not board_d.is_dir() or board_d.name.startswith("."):
                continue

            layout_conf = self.parse_layout_conf(board_d)
            repo_name = layout_conf.get("repo-name")

            # e.g. overlay-amd64-host doesn't have layout.conf
            if repo_name is None:
                repo_name = self.read_profiles_repo_name(board_d)

            if repo_name is None:
                self.logger.warning(
                    "Couldn't find a canonical name for board dir '{}'."
                    .format(board_d.name)
                )
                repo_name = board_d.name

            repo_names[board_d.name] = repo_name
            board_relations.add_node(repo_name)

        for overlay, repo_name in repo_names.items():
            board_d = self.board_overlays_repo / overlay

            for parent in self.get_profiles_base_parent_boards(board_d):
                if parent != repo_name:
                    board_relations.add_edge(parent, repo_name)

            # Various model/skus of recent boards don't have explicit overlay
            # dirs, but are specified in model.yaml in the base overlay
            for child in self.get_model_yaml_boards(board_d):
                if repo_name != child:
                    board_relations.add_edge(repo_name, child)

            # Some relations only exists in layout.conf, e.g.
            # - x86-generic -> x86-generic_embedded
            # - project-* dirs and their children
            # - peach -> peach_pit in firmware-gru-8785.B
            layout_conf = self.parse_layout_conf(board_d)
            for parent in layout_conf.get("masters", "").split():
                if parent != repo_name and parent not in (
                    "chromiumos",
                    "portage-stable",
                    "eclass-overlay",
                ):
                    board_relations.add_edge(parent, repo_name)

        # "snow" is the default, implicit "daisy"
        if board_relations.nodes().intersection(("snow", "daisy")):
            board_relations.add_edge("daisy", "snow")

        # Some newer board variants are only in this project repo
        for board in iterdir(self.chromiumos_project_repo):
            if not board.is_dir() or board.name.startswith("."):
                continue

            board_relations.add_node(board.name)

            for profile in iterdir(board):
                if not profile.is_dir() or profile.name.startswith("."):
                    continue

                # puff/puff exists
                if profile.name != board.name:
                    board_relations.add_edge(board.name, profile.name)

                for child in self.get_project_config_boards(profile):
                    # shadowkeep/shadowkeep/shadowkeep exists
                    if child == profile.name == board.name:
                        continue

                    # galaxy/{andromeda,sombrero} has galaxy
                    # make them {andromeda,sombrero}_galaxy
                    elif child == board.name:
                        child = "{}_{}".format(profile.name, child)

                    if child != profile.name:
                        board_relations.add_edge(profile.name, child)

        # Project repo lists all "veyron" boards under "veyron-pinky"
        if "veyron-pinky" in board_relations.nodes():
            board_relations.add_edge("veyron", "veyron-pinky")
            for child in board_relations.children("veyron-pinky"):
                board_relations.add_edge("veyron", child)
                board_relations.remove_edge("veyron-pinky", child)

        # Weird stuff from depthcharge
        for board, config in self.depthcharge_boards.items():
            parent = config.get("BOARD", None)
            parent = config.get("BOARD_DIR", parent)
            if parent is None:
                continue

            board_relations.add_node(board)

            # src/board/ and BOARD_DIR has gru (baseboard) and veyron_*
            # (variants), we can't just always add "baseboard-".
            if "baseboard-{}".format(parent) in board_relations.nodes():
                parent = "baseboard-{}".format(parent)

            # This looks incorrect for a few boards, so only add the
            # relation if we don't know anything about the board
            if not board_relations.parents(board):
                if parent != board:
                    board_relations.add_edge(parent, board)

        nodes = {
            node.replace("-", "_"): node
            for node in board_relations.nodes()
        }

        def coreboot_board_name(config):
            if config is None or not config.startswith("BOARD_"):
                return None

            board = "_".join(config.split("_")[2:]).lower()
            if board.startswith("baseboard_"):
                board = "baseboard-{}".format(board[len("baseboard_"):])

            if board not in self.coreboot_boards:
                return None

            return board

        def add_coreboot_parents(board):
            if board is None:
                return None

            board = nodes.get(board.replace("-", "_"), board)
            board_relations.add_node(board)

            block = self.coreboot_boards.get(board, {})
            parents = set(
                coreboot_board_name(config)
                for config, value in block.items()
                if value
            )

            for parent in parents - {board, None}:
                add_coreboot_parents(parent)
                parent = nodes.get(parent.replace("-", "_"), parent)

                # This also has conflicts with board-overlays
                if not board_relations.parents(board):
                    board_relations.add_edge(parent, board)

        for board, block in self.coreboot_boards.items():
            add_coreboot_parents(board)

        nodes = {
            node.replace("_", "-"): node
            for node in board_relations.nodes()
        }

        # Recovery.conf heuristics, doesn't have actual parent board info
        for board, blocks in self.recovery_conf_boards.items():
            parents = set([b.get("file").split("_")[2] for b in blocks])
            parents.discard(board)
            if len(parents) > 1:
                continue
            elif len(parents) == 0:
                parent = None
            else:
                parent = parents.pop()

            # This is really inaccurate with underscores replaced with
            # hyphens, so only use it if we don't know anything else
            if board in nodes:
                continue

            # Don't duplicate veyron_speedy as speedy
            if parent in nodes:
                parent = nodes[parent]
            if parent and parent.endswith(board):
                parent = parent[:-len(board)-1]

            board_relations.add_node(board)
            if parent:
                board_relations.add_edge(parent, board)

        # Add board architectures as root parent
        for board, config in self.depthcharge_boards.items():
            if config.get("ARCH_X86"):
                arch = "amd64"
            elif config.get("ARCH_ARM_V8"):
                arch = "arm64"
            elif config.get("ARCH_ARM"):
                arch = "arm"
            else:
                continue

            roots = board_relations.roots(board)
            for root in roots - {"x86", "amd64", "arm64", "arm"}:
                board_relations.add_edge(arch, root)

        # Baseboards, chipsets shouldn't depend on others in their class
        for board in board_relations.nodes():
            if board.startswith("chipset-"):
                for child in board_relations.children(board):
                    if child.startswith("chipset-"):
                        board_relations.remove_edge(board, child)
                        for parent in board_relations.parents(board):
                            board_relations.add_edge(parent, child)

            elif board.startswith("baseboard-"):
                for child in board_relations.children(board):
                    if child.startswith("baseboard-"):
                        board_relations.remove_edge(board, child)
                        for parent in board_relations.parents(board):
                            board_relations.add_edge(parent, child)

        # Relations from older versions no longer in main branches,
        # coreboot Kconfigs I'm too lazy to parse, etc.
        for parent, children in [
            ("amd64", ["chipset-pinetrail", "chipset-snb", "chipset-ivb",
                       "chipset-hsw", "chipset-cnl", "chipset-icl",
                       "chipset-rpl", "reven"]),
            ("chipset-adl", ["adlrvp", "shadowmountain"]),
            ("chipset-bdw", ["baseboard-auron"]),
            ("chipset-cml", ["cmlrvp"]),
            ("chipset-cnl", ["cnlrvp"]),
            ("chipset-glk", ["glkrvp"]),
            ("chipset-hsw", ["baseboard-slippy", "baseboard-beltino"]),
            ("chipset-icl", ["iclrvp", "dragonegg"]),
            ("chipset-ivb", ["stout"]),
            ("chipset-kbl", ["kblrvp"]),
            ("chipset-pinetrail", ["x86-alex-he", "x86-mario", "x86-zgb-he"]),
            ("chipset-snb", ["butterfly", "lumpy", "stumpy"]),
            ("chipset-tgl", ["tglrvp", "deltaur", "deltan", ]),
            ("chipset-whl", ["whlrvp"]),
            ("chipset-mendocino", ["chausie"]),
            ("baseboard-auron", ["auron"]),
            ("baseboard-slippy", ["slippy"]),
            ("baseboard-beltino", ["beltino"]),
            ("arm", ["chipset-tegra124", "chipset-exynos5", "chipset-cygnus",
                     "chipset-ipq4019", "chipset-ipq8064", "chipset-rk3288"]),
            ("chipset-cygnus", ["purin"]),
            ("chipset-exynos5", ["daisy", "peach"]),
            ("chipset-ipq4019", ["gale"]),
            ("chipset-ipq8064", ["storm"]),
            ("chipset-rk3288", ["veyron"]),
            ("chipset-tegra124", ["nyan"]),
            ("storm", ["arkham", "whirlwind"]),
            ("veyron", ["veyron_mickey", "veyron_rialto"]),
            ("arm64", ["chipset-tegra210", "chipset-qcs404", "chipset-mt8188g"]),
            ("chipset-qcs404", ["mistral"]),
            ("chipset-tegra210", ["foster", "smaug"]),
        ]:
            for child in children:
                board_relations.add_edge(parent, child)

        return board_relations

    @options.add
    @Argument("-o", "--output", required=True)
    def output(self, path):
        """Write updated config to PATH."""

        if path is None:
            raise ValueError(
                "Output argument is required."
            )

        return Path(path).resolve()

    @property
    @lru_cache
    def board_config_sections(self):
        board_relations = self.board_relations

        # "project-*" overlays don't really look like boards.
        if self.board_overlays_repo is not None:
            projects = set(
                overlay.name.partition("-")[2]
                for overlay in self.board_overlays_repo.glob("project-*")
            )
        else:
            projects = set()

        nonboards = set((
            *projects,
            "unprovisioned",
            "signed",
            "embedded",
            "legacy",
            "npcx796",
            "npcx796fc",
            "ext_ec",
            "extec",
            "alc1015_amp",
        ))

        chipsets = {
            "chipset-adl": "alderlake",
            "chipset-adln": "alderlake-n",
            "chipset-apl": "apollolake",
            "chipset-bdw": "broadwell",
            "chipset-bsw": "braswell",
            "chipset-byt": "baytrail",
            "chipset-cml": "cometlake",
            "chipset-cnl": "cannonlake",
            "chipset-glk": "geminilake",
            "chipset-hsw": "haswell",
            "chipset-icl": "icelake",
            "chipset-ivb": "ivybridge",
            "chipset-jsl": "jasperlake",
            "chipset-kbl": "kabylake",
            "chipset-mtl": "meteorlake",
            "chipset-rpl": "raptorlake",
            "chipset-skl": "skylake",
            "chipset-snb": "sandybridge",
            "chipset-tgl": "tigerlake",
            "chipset-whl": "whiskeylake",
            "chipset-stnyridge": "stoneyridge",
        }

        def get_parent(board):
            # Projects can be the sole parent of actual boards (e.g.
            # freon was to a lot of boards) so don't use them as parents
            # at all, despite breaking e.g. termina/tael parentage.
            parents = board_relations.parents(board) - nonboards
            if len(parents) > 1:
                self.logger.warning(
                    "Board '{}' has multiple parents: '{}'"
                    .format(board, parents)
                )
            elif len(parents) == 0:
                return None

            # Prefer longer chains
            return max(
                parents,
                key=lambda p: len(board_relations.ancestors(p)),
            )

        aliases = {}
        def add_alias(alias, board):
            if alias in aliases:
                aliases[alias] = None
            else:
                aliases[alias] = board

        # Do not alias nonboards to anything
        for nonboard in nonboards:
            aliases[nonboard] = None

        # Convert the nodes to the path-to-node format we want
        paths = {}
        for board in board_relations.nodes():
            parts = [board]
            parent = get_parent(board)

            # Prefer full names for chipsets
            if board.startswith("chipset-"):
                chipset = chipsets.get(board, board[len("chipset-"):])
                parts = [chipset]

            # Don't keep baseboard prefix
            if board.startswith("baseboard-"):
                baseboard = board[len("baseboard-"):]
                parts = [baseboard]

            if parent is not None:
                lhs, sep, rhs = board.partition("_")

                if sep != "_":
                    pass

                # Fixup left-duplication e.g. veyron/veyron_speedy
                elif lhs == parent:
                    parts = [rhs]
                    add_alias(rhs, board)

                # Fixup right-duplication e.g. hatch/unprovisioned_hatch
                elif rhs == parent:
                    parts = [lhs]
                    add_alias(lhs, board)

                # Split e.g. unprovisioned_kohaku -> kohaku/unprovisioned
                elif lhs in nonboards:
                    parts = [lhs, rhs]

                # e.g. arcada_signed, volteer2_ti50, helios_diskswap etc.
                else:
                    parts = [rhs, lhs]

            while parent is not None:
                # Prefer full names for chipsets
                if parent.startswith("chipset-"):
                    chipset = chipsets.get(parent, parent[len("chipset-"):])
                    parts.append(chipset)

                # Normalize boards with the same name as baseboard
                elif parent.startswith("baseboard-"):
                    baseboard = parent[len("baseboard-"):]
                    if parts[-1] != baseboard:
                        parts.append(baseboard)

                else:
                    parts.append(parent)

                parent = get_parent(parent)

            paths[board] = "boards/{}".format("/".join(reversed(parts)))

        for alias, board in aliases.items():
            if board is not None:
                paths.setdefault(alias, paths[board])

        return paths

    def __call__(self):
        config = configparser.ConfigParser(
            dict_type=SortedDict(lambda s: s.split('/')),
        )

        for arch in ("x86", "amd64", "arm", "arm64"):
            name = self.board_config_sections.get(arch, None)
            if name is None:
                continue

            config.add_section(name)
            config[name]["arch"] = arch
            config[name]["codename"] = "{}-generic".format(arch)

        for board, name in self.board_config_sections.items():
            if board.startswith("chipset-"):
                config.add_section(name)
                config[name]["codename"] = board

        for codename, blocks in self.recovery_conf_boards.items():
            name = self.board_config_sections.get(codename, None)
            if name is None:
                continue

            config.add_section(name)
            board = config[name]
            board["codename"] = codename

            for i, block in enumerate(blocks):
                if len(blocks) > 1:
                    name_i = "{}/{}".format(name, i)
                    config.add_section(name_i)
                    board = config[name_i]

                if block.get("hwidmatch", None):
                    board["hwid-match"] = block["hwidmatch"]

                if block.get("name", None):
                    board["name"] = block["name"]

        # Some heuristics for kernel compression
        if self.depthcharge_repo is not None:
            arm64_boot_c = (self.depthcharge_repo / "src/arch/arm/boot64.c")
            if arm64_boot_c.is_file():
                arm64_boot_c = arm64_boot_c.read_text()
            else:
                arm64_boot_c = ""

            fit_c = (self.depthcharge_repo / "src/boot/fit.c")
            if fit_c.is_file():
                fit_c = fit_c.read_text()
            else:
                fit_c = ""

            fit_h = (self.depthcharge_repo / "src/boot/fit.h")
            if fit_h.is_file():
                fit_h = fit_h.read_text()
            else:
                fit_h = ""

        else:
            arm64_boot_c = ""
            fit_c = ""
            fit_h = ""

        if "fit_decompress(kernel" in arm64_boot_c:
            arm64_lz4_kernel = "CompressionLz4" in fit_h + fit_c
            arm64_lzma_kernel = "CompressionLzma" in fit_h + fit_c
        elif "switch(kernel->compression)" in arm64_boot_c:
            arm64_lz4_kernel = "case CompressionLz4" in arm64_boot_c
            arm64_lzma_kernel = "case CompressionLzma" in arm64_boot_c
        else:
            arm64_lz4_kernel = False
            arm64_lzma_kernel = False

        for codename, block in self.depthcharge_boards.items():
            name = self.board_config_sections.get(codename, None)
            if name is None:
                name = codename
            if name not in config:
                config.add_section(name)

            board = config[name]
            board["codename"] = codename

            if block.get("KERNEL_SIZE", None):
                board["image-max-size"] = str(block["KERNEL_SIZE"])

            if block.get("KERNEL_FIT", False):
                board["image-format"] = "fit"
            elif block.get("KERNEL_ZIMAGE", False):
                board["image-format"] = "zimage"

            if block.get("ARCH_ARM_V8", False):
                board["boots-lz4-kernel"] = str(arm64_lz4_kernel)
                board["boots-lzma-kernel"] = str(arm64_lzma_kernel)

            # compatible string
            board_c = (
                self.depthcharge_repo / "src/board"
                / codename / "board.c"
            )
            board_c = board_c.read_text() if board_c.is_file() else ""
            if block.get("KERNEL_FIT", False):
                m = re.search(
                    'fit_(?:add|set)_compat(?:_by_rev)?\('
                    '"([^"]+?)(?:-rev[^-]+|-sku[^-]+)*"',
                    board_c,
                )
                if m:
                    board["dt-compatible"] = m.group(1)

                elif "sprintf(compat, pattern, CONFIG_BOARD," in fit_c:
                    board["dt-compatible"] = "google,{}".format(
                        block.get("BOARD", codename).lower()
                        .replace("_", "-").replace(" ", "-")
                    )

                elif '"google,%s", mb_part_string' in fit_c:
                    block = self.coreboot_boards.get(codename, {})
                    mb_part_string = block.get(
                        "MAINBOARD_PART_NUMBER",
                        codename,
                    )
                    board["dt-compatible"] = "google,{}".format(
                        mb_part_string.lower()
                        .replace("_", "-").replace(" ", "-")
                    )

        def graph(config):
            graph = DirectedGraph()
            for section in config.sections():
                parts = section.split('/')
                for i, part in enumerate(parts):
                    parent = str.join("/", parts[:i])
                    child = str.join("/", (*parts[:i], part))
                    graph.add_node(child)
                    if parent:
                        graph.add_edge(parent, child)
            return graph

        # Propagate config values upwards
        def propagate_configs(graph, config, section, keys):
            commons = {}

            for child in graph.children(section):
                propagate_configs(graph, config, child, keys)

            if section not in config:
                config.add_section(section)

            for k in keys:
                values = set(
                    config[child].get(k, None)
                    for child in graph.children(section)
                )
                if len(values) == 1:
                    value = values.pop()
                    if value is not None:
                        config[section].setdefault(k, value)

            for child in graph.children(section):
                for k in keys:
                    if k in config[section] and k in config[child]:
                        if config[section][k] == config[child][k]:
                            config[child].pop(k)

        # Don't propagate up to the root, only up to baseboard nodes.
        g = graph(config)
        for arch in g.children("boards"):
            for chipset in g.children(arch):
                for baseboard in g.children(chipset):
                    propagate_configs(g, config, baseboard, {
                        "image-format", "image-max-size",
                        "boots-lz4-kernel", "boots-lzma-kernel",
                    })

        # Trim unreleased boards that don't have names, hwid-matches
        max_depth = max(b.count('/') for b in config.sections())
        for _ in range(0, max_depth):
            deleted = False

            for section in graph(config).leaves():
                c = config[section]
                if "hwid-match" not in c and "name" not in c:
                    self.logger.warning(
                        "Skipping unreleased board '{}'."
                        .format(section)
                    )
                    del config[section]
                    deleted = True

            if not deleted:
                break

        with self.output.open("w") as output_f:
            config.write(output_f)

if __name__ == "__main__":
    update_config.main()



#! /usr/bin/env python3
# SPDX-License-Identifier: GPL-2.0-or-later

# depthcharge-tools depthchargectl write subcommand
# Copyright (C) 2020-2022 Alper Nebi Yasak <alpernebiyasak@gmail.com>
# See COPYRIGHT and LICENSE files for full copyright information.

import argparse
import logging
import os
import subprocess

from pathlib import Path

from depthcharge_tools import (
    __version__,
)
from depthcharge_tools.utils.argparse import (
    Command,
    Argument,
    Group,
    CommandExit,
)
from depthcharge_tools.utils.platform import (
    KernelEntry,
    installed_kernels,
)

from depthcharge_tools.depthchargectl import depthchargectl


class ImageBuildError(CommandExit):
    def __init__(self, kernel_version=None):
        self.kernel_version = kernel_version

        if kernel_version is None:
            message = "Failed to build depthcharge image."

        else:
            message = (
                "Failed to build depthcharge image for kernel version '{}'."
                .format(kernel_version)
            )

        super().__init__(message=message)


class NotBootableImageError(CommandExit):
    def __init__(self, image):
        self.image = image
        super().__init__(
            "Image '{}' is not bootable on this board."
            .format(image)
        )


class NoUsableCrosPartitionError(CommandExit):
    def __init__(self):
        super().__init__(
            "No usable Chrome OS Kernel partition found."
        )


@depthchargectl.subcommand("write")
class depthchargectl_write(
    depthchargectl,
    prog="depthchargectl write",
    usage="%(prog)s [options] [KERNEL-VERSION | IMAGE]",
    add_help=False,
):
    """Write an image to a ChromeOS kernel partition."""

    _logger = depthchargectl._logger.getChild("write")
    config_section = "depthchargectl/write"

    @depthchargectl.board.copy()
    def board(self, codename=""):
        """Assume we're running on the specified board"""
        # We can write images to partitions without knowing the board.
        # The image argument will become required if this returns None.
        try:
            return super().board
        except Exception as err:
            self.logger.warning(err)
            return None

    @Group
    def positionals(self):
        """Positional arguments"""

        if self.image is not None and self.kernel_version is not None:
            raise ValueError(
                "Image and kernel_version arguments are mutually exclusive"
            )

        arg = self.image or self.kernel_version

        # Turn arg into a relevant KernelEntry if it's a kernel version
        # or a Path() if not
        if isinstance(arg, str):
            arg = max(
                (k for k in installed_kernels() if k.release == arg),
                default=Path(arg).resolve(),
            )

        if isinstance(arg, KernelEntry):
            self.image = None
            self.kernel_version = arg

        elif isinstance(arg, Path):
            self.image = arg
            self.kernel_version = None

        if self.board is None and self.image is None:
            raise ValueError(
                "An image file is required when no board is specified."
            )

    @positionals.add
    @Argument(dest=argparse.SUPPRESS, nargs=0)
    def kernel_version(self, kernel_version):
        """Installed kernel version to write to disk."""
        return kernel_version

    @positionals.add
    @Argument
    def image(self, image=None):
        """Depthcharge image to write to disk."""
        return image

    @Group
    def options(self):
        """Options"""

    @options.add
    @Argument("-f", "--force", force=True)
    def force(self, force=False):
        """Write image even if it cannot be verified."""
        return force

    @options.add
    @Argument("-t", "--target", metavar="DISK|PART")
    def target(self, target):
        """Specify a disk or partition to write to."""
        return target

    @options.add
    @Argument("--no-prioritize", prioritize=False)
    def prioritize(self, prioritize=True):
        """Don't set any flags on the partition."""
        return prioritize

    @options.add
    @Argument("--allow-current", allow=True)
    def allow_current(self, allow=False):
        """Allow overwriting the currently booted partition."""
        return allow

    def __call__(self):
        if self.board is None:
            self.logger.warning(
                "Using given image '{}' without board-specific checks."
                .format(self.image)
            )
            image = self.image

        elif self.image is not None:
            self.logger.info("Using given image '{}'." .format(self.image))
            image = self.image

            try:
                depthchargectl.check(
                    image=image,
                    config=self.config,
                    board=self.board,
                    tmpdir=self.tmpdir / "check",
                    images_dir=self.images_dir,
                    vboot_keyblock=self.vboot_keyblock,
                    vboot_public_key=self.vboot_public_key,
                    vboot_private_key=self.vboot_private_key,
                    kernel_cmdline=self.kernel_cmdline,
                    ignore_initramfs=self.ignore_initramfs,
                    verbosity=self.verbosity,
                )

            except Exception as err:
                if self.force:
                    self.logger.warning(
                        "Image '{}' is not bootable on this board, "
                        "continuing due to --force."
                        .format(image)
                    )

                else:
                    raise NotBootableImageError(image) from err

        else:
            # No image given, try creating one.
            try:
                image = depthchargectl.build_(
                    kernel_version=self.kernel_version,
                    root=self.root,
                    root_mountpoint=self.root_mountpoint,
                    boot_mountpoint=self.boot_mountpoint,
                    config=self.config,
                    board=self.board,
                    tmpdir=self.tmpdir / "build",
                    images_dir=self.images_dir,
                    vboot_keyblock=self.vboot_keyblock,
                    vboot_public_key=self.vboot_public_key,
                    vboot_private_key=self.vboot_private_key,
                    kernel_cmdline=self.kernel_cmdline,
                    ignore_initramfs=self.ignore_initramfs,
                    verbosity=self.verbosity,
                )

            except Exception as err:
                raise ImageBuildError(self.kernel_version) from err

        # We don't want target to unconditionally avoid the current
        # partition since we will also check that here. But whatever we
        # choose must be bigger than the image we'll write to it.
        self.logger.info("Searching disks for a target partition.")
        try:
            target = depthchargectl.target(
                disks=[self.target] if self.target else [],
                min_size=image.stat().st_size,
                allow_current=self.allow_current,
                root=self.root,
                root_mountpoint=self.root_mountpoint,
                boot_mountpoint=self.boot_mountpoint,
                config=self.config,
                board=self.board,
                tmpdir=self.tmpdir / "target",
                images_dir=self.images_dir,
                vboot_keyblock=self.vboot_keyblock,
                vboot_public_key=self.vboot_public_key,
                vboot_private_key=self.vboot_private_key,
                kernel_cmdline=self.kernel_cmdline,
                ignore_initramfs=self.ignore_initramfs,
                verbosity=self.verbosity,
            )

        except Exception as err:
            raise NoUsableCrosPartitionError() from err

        if target is None:
            raise NoUsableCrosPartitionError()

        self.logger.info("Targeted partition '{}'.".format(target))

        # Check and warn if we targeted the currently booted partition,
        # as that usually means it's the only partition.
        current = self.diskinfo.by_kern_guid()
        if current is not None and self.allow_current and target.path == current.path:
            self.logger.warning(
                "Overwriting the currently booted partition '{}'. "
                "This might make your system unbootable."
                .format(target)
            )

        self.logger.info(
            "Writing image '{}' to partition '{}'."
            .format(image, target)
        )
        target.write_bytes(image.read_bytes())
        self.logger.warning(
            "Wrote image '{}' to partition '{}'."
            .format(image, target)
        )

        if self.prioritize:
            self.logger.info(
                "Setting '{}' as the highest-priority bootable part."
                .format(target)
            )
            try:
                depthchargectl.bless(
                    partition=target,
                    oneshot=True,
                    root=self.root,
                    root_mountpoint=self.root_mountpoint,
                    boot_mountpoint=self.boot_mountpoint,
                    config=self.config,
                    board=self.board,
                    tmpdir=self.tmpdir / "bless",
                    images_dir=self.images_dir,
                    vboot_keyblock=self.vboot_keyblock,
                    vboot_public_key=self.vboot_public_key,
                    vboot_private_key=self.vboot_private_key,
                    kernel_cmdline=self.kernel_cmdline,
                    ignore_initramfs=self.ignore_initramfs,
                    verbosity=self.verbosity,
                )
            except Exception as err:
                raise CommandExit(
                    "Failed to set '{}' as the highest-priority bootable part."
                    .format(target)
                ) from err

            self.logger.warning(
                "Set partition '{}' as next to boot."
                .format(target)
            )

        return target

    global_options = depthchargectl.global_options
    config_options = depthchargectl.config_options




#! /usr/bin/env python3
# SPDX-License-Identifier: GPL-2.0-or-later

# depthcharge-tools depthchargectl list subcommand
# Copyright (C) 2020-2022 Alper Nebi Yasak <alpernebiyasak@gmail.com>
# See COPYRIGHT and LICENSE files for full copyright information.

import argparse
import logging
import subprocess

from depthcharge_tools import __version__
from depthcharge_tools.utils.argparse import (
    Command,
    Argument,
    Group,
    CommandExit,
)
from depthcharge_tools.utils.collections import (
    TypedList
)
from depthcharge_tools.utils.os import (
    Disk,
    CrosPartition,
)

from depthcharge_tools.depthchargectl import depthchargectl


class CrosPartitions(TypedList(CrosPartition)):
    def __init__(self, partitions=None, columns=None, headings=True):
        super().__init__(partitions)

        if columns is None:
            if any(part.path is None for part in partitions):
                columns = ["S", "P", "T", "DISKPATH", "PARTNO"]
            else:
                columns = ["S", "P", "T", "PATH"]

        self._headings = headings
        self._columns = columns

    def _row(self, part):
        values = {}

        if set(self._columns).intersection((
            "A", "S", "P", "T",
            "ATTRIBUTE", "SUCCESSFUL", "PRIORITY", "TRIES",
        )):
            flags = part.flags
            values.update({
                "A": flags["attribute"],
                "S": flags["successful"],
                "P": flags["priority"],
                "T": flags["tries"],
                "ATTRIBUTE": flags["attribute"],
                "SUCCESSFUL": flags["successful"],
                "PRIORITY": flags["priority"],
                "TRIES": flags["tries"],
            })

        if "SIZE" in self._columns:
            values["SIZE"] = part.size

        if part.path is not None:
            values["PATH"] = part.path

        if part.disk is not None and part.disk.path is not None:
            values["DISK"] = part.disk.path
            values["DISKPATH"] = part.disk.path

        if part.partno is not None:
            values["PARTNO"] = part.partno

        return [str(values.get(c, "")) for c in self._columns]

    def __str__(self):
        rows = []

        if self._headings:
            rows.append(self._columns)

        parts = sorted(self, key=lambda p: p.path or p.disk.path)
        rows.extend(self._row(part) for part in parts)

        # Using tab characters makes things misalign when the data
        # widths vary, so find max width for each column from its data,
        # and format everything to those widths.
        widths = [max(4, *map(len, col)) for col in zip(*rows)]
        fmt = " ".join("{{:{w}}}".format(w=w) for w in widths)
        return "\n".join(fmt.format(*row) for row in rows)


@depthchargectl.subcommand("list")
class depthchargectl_list(
    depthchargectl,
    prog="depthchargectl list",
    usage="%(prog)s [options] [DISK ...]",
    add_help=False,
):
    """List ChromeOS kernel partitions."""

    _logger = depthchargectl._logger.getChild("list")
    config_section = "depthchargectl/list"

    @depthchargectl.board.copy()
    def board(self, codename=""):
        """Assume we're running on the specified board"""
        # We can list partitions without knowing the board.
        try:
            return super().board
        except Exception as err:
            self.logger.warning(err)
            return None

    @Group
    def positionals(self):
        """Positional arguments"""

    @positionals.add
    @Argument(metavar="DISK")
    def disks(self, *disks):
        """Disks to check for ChromeOS kernel partitions."""

        if self.all_disks:
            self.logger.info("Searching all disks.")
            disks = self.diskinfo.roots()
        elif disks:
            self.logger.info(
                "Searching real disks for {}."
                .format(", ".join(str(d) for d in disks))
            )
            images = []
            for d in disks:
                if self.diskinfo.evaluate(d) is None:
                    try:
                        images.append(Disk(d))
                    except ValueError as err:
                        self.logger.warning(
                            err,
                            exc_info=self.logger.isEnabledFor(logging.DEBUG),
                        )
            disks = [*self.diskinfo.roots(*disks), *images]
        else:
            self.logger.info("Searching bootable disks.")
            root = (
                self.diskinfo.by_mountpoint("/", fstab_only=True)
                or self.diskinfo.by_mountpoint(self.root_mountpoint)
            )
            boot = (
                self.diskinfo.by_mountpoint("/boot", fstab_only=True)
                or self.diskinfo.by_mountpoint(self.boot_mountpoint)
            )
            disks = self.diskinfo.roots(root, boot)

        if disks:
            self.logger.info(
                "Using disks: {}."
                .format(", ".join(str(d) for d in disks))
            )
        else:
            raise ValueError("Could not find any matching disks.")

        return disks

    @Group
    def options(self):
        """Options"""
        if self.count and self.output:
            raise ValueError(
                "Count and Output arguments are mutually exclusive."
            )

    @options.add
    @Argument("-n", "--noheadings", headings=False)
    def headings(self, headings=True):
        """Don't print column headings."""
        return headings

    @options.add
    @Argument("-a", "--all-disks", all_disks=True)
    def all_disks(self, all_disks=False):
        """List partitions on all disks."""
        return all_disks

    valid_columns = {
        "ATTRIBUTE", "SUCCESSFUL", "PRIORITY", "TRIES",
        "A", "S", "P", "T",
        "PATH",
        "DISKPATH", "DISK",
        "PARTNO",
        "SIZE",
    }

    @options.add
    @Argument("-o", "--output", nargs=1, append=True)
    def output(self, *columns):
        """Comma separated list of columns to output."""

        if len(columns) == 0:
            self.logger.info("Using default output format.")
            return None

        elif len(columns) == 1 and isinstance(columns[0], str):
            columns = columns[0]
            self.logger.info("Using output format '{}'.".format(columns))

        else:
            columns = ",".join(columns)
            self.logger.info("Using output format '{}'.".format(columns))

        columns = columns.split(',')

        invalid_columns = sorted(
            set(columns).difference(self.valid_columns),
            key=columns.index,
        )
        if invalid_columns:
            raise ValueError(
                "Unsupported output columns '{}'."
                .format(invalid_columns)
            )

        return columns

    @options.add
    @Argument("-c", "--count", count=True)
    def count(self, count=False):
        """Print only the count of partitions."""
        return count

    def __call__(self):
        parts = []
        error_disks = []

        for disk in self.disks:
            try:
                parts.extend(disk.cros_partitions())
            except subprocess.CalledProcessError as err:
                error_disks.append(disk)
                self.logger.debug(
                    "Couldn't get partitions for disk '{}'."
                    .format(disk)
                )
                self.logger.debug(
                    err,
                    exc_info=self.logger.isEnabledFor(logging.DEBUG),
                )

        if self.count:
            output = len(parts)

        else:
            output = CrosPartitions(
                parts,
                headings=self.headings,
                columns=self.output,
            )

        if error_disks:
            return CommandExit(
                message=(
                    "Couldn't get partitions for disks {}."
                    .format(", ".join(str(d) for d in error_disks))
                ),
                output=output,
                returncode=1,
            )

        return output

    global_options = depthchargectl.global_options
    config_options = depthchargectl.config_options




#! /usr/bin/env python3
# SPDX-License-Identifier: GPL-2.0-or-later

# depthcharge-tools depthchargectl config subcommand
# Copyright (C) 2021-2022 Alper Nebi Yasak <alpernebiyasak@gmail.com>
# See COPYRIGHT and LICENSE files for full copyright information.

import argparse
import logging

from depthcharge_tools import __version__
from depthcharge_tools.utils.argparse import (
    Command,
    Argument,
    Group,
    CommandExit,
)

from depthcharge_tools.depthchargectl import depthchargectl


@depthchargectl.subcommand("config")
class depthchargectl_config(
    depthchargectl,
    prog="depthchargectl config",
    usage="%(prog)s [options] KEY",
    add_help=False,
):
    """Get depthchargectl configuration values."""

    _logger = depthchargectl._logger.getChild("config")
    config_section = "depthchargectl/config"

    @depthchargectl.board.copy()
    def board(self, codename=""):
        """Assume we're running on the specified board"""
        # We can query configs without knowing the board.
        try:
            return super().board
        except Exception as err:
            self.logger.warning(err)
            return None

    @Group
    def positionals(self):
        """Positional arguments"""

    @positionals.add
    @Argument
    def key(self, key):
        """Config key to get value of."""
        return key

    @Group
    def options(self):
        """Options"""

    @options.add
    @Argument("--section", nargs=1)
    def section(self, section=None):
        """Config section to work on."""
        parser = self.config.parser

        if section is None:
            section = self.config.name

        if section not in parser.sections():
            if section != parser.default_section:
                parser.add_section(section)

        return parser[section]

    @options.add
    @Argument("--default", nargs=1)
    def default(self, default=None):
        """Value to return if key doesn't exist in section."""
        return default

    def __call__(self):
        if self.key not in self.section:
            if self.default is not None:
                return self.default
            else:
                raise KeyError(
                    "Key '{}' not found in section '{}'."
                    .format(self.key, self.section.name)
                )

        return self.section[self.key]

    global_options = depthchargectl.global_options
    config_options = depthchargectl.config_options



#! /usr/bin/env python3
# SPDX-License-Identifier: GPL-2.0-or-later

# depthcharge-tools depthchargectl bless subcommand
# Copyright (C) 2020-2022 Alper Nebi Yasak <alpernebiyasak@gmail.com>
# See COPYRIGHT and LICENSE files for full copyright information.

import argparse
import logging
import subprocess

from depthcharge_tools import __version__
from depthcharge_tools.utils.argparse import (
    Command,
    Argument,
    Group,
    CommandExit,
)
from depthcharge_tools.utils.os import (
    Disk,
    Partition,
    CrosPartition,
)
from depthcharge_tools.utils.platform import (
    is_cros_boot,
)

from depthcharge_tools.depthchargectl import depthchargectl


@depthchargectl.subcommand("bless")
class depthchargectl_bless(
    depthchargectl,
    prog="depthchargectl bless",
    usage="%(prog)s [options] [DISK | PARTITION]",
    add_help=False,
):
    """Set the active or given partition as successfully booted."""

    _logger = depthchargectl._logger.getChild("bless")
    config_section = "depthchargectl/bless"

    @depthchargectl.board.copy()
    def board(self, codename=""):
        """Assume we're running on the specified board"""
        # We can bless partitions without knowing the board.
        try:
            return super().board
        except Exception as err:
            self.logger.warning(err)
            return None

    @Group
    def positionals(self):
        """Positional arguments"""
        if self.disk is not None and self.partition is not None:
            raise ValueError(
                "Disk and partition arguments are mutually exclusive."
            )

        device = self.disk or self.partition

        if isinstance(device, str):
            sys_device = self.diskinfo.evaluate(device)

            if sys_device is not None:
                self.logger.info(
                    "Using argument '{}' as a block device."
                    .format(device)
                )
                device = sys_device

            else:
                self.logger.info(
                    "Using argument '{}' as a disk image."
                    .format(device)
                )
                device = Disk(device)

        if isinstance(device, Disk):
            if self.partno is None:
                raise ValueError(
                    "Partno argument is required for disks."
                )
            partition = device.partition(self.partno)

        elif isinstance(device, Partition):
            if self.partno is not None and self.partno != device.partno:
                raise ValueError(
                    "Partition and partno arguments are mutually exclusive."
                )
            partition = device

        elif device is None:
            self.logger.info(
                "No partition given, defaulting to currently booted one."
            )
            partition = self.diskinfo.by_kern_guid()

        if partition is None:
            if is_cros_boot():
                raise ValueError(
                    "Couldn't figure out the currently booted partition."
                )
            else:
                raise ValueError(
                    "A disk or partition argument is required when not "
                    "booted with depthcharge."
                )

        self.logger.info(
            "Working on partition '{}'."
            .format(partition)
        )

        try:
            cros_partitions = partition.disk.cros_partitions()
        except subprocess.CalledProcessError as err:
            self.logger.debug(
                err,
                exc_info=self.logger.isEnabledFor(logging.DEBUG),
            )
            raise ValueError(
                "Couldn't get partitions for disk '{}'."
                .format(partition.disk)
            ) from err

        if partition not in cros_partitions:
            raise ValueError(
                "Partition '{}' is not a ChromeOS Kernel partition"
                .format(partition)
            )

        partition = CrosPartition(partition)
        self.partition = partition
        self.disk = partition.disk
        self.partno = partition.partno

    @positionals.add
    @Argument(nargs=0)
    def disk(self, disk=None):
        """Disk image to manage partitions of"""
        return disk

    @positionals.add
    @Argument
    def partition(self, partition=None):
        """ChromeOS Kernel partition device to manage"""
        return partition

    @Group
    def options(self):
        """Options"""

    @options.add
    @Argument("-i", "--partno", nargs=1)
    def partno(self, number=None):
        """Partition number in the given disk image"""
        try:
            if number is not None:
                number = int(number)
        except:
            raise TypeError(
                "Partition number must be a positive integer."
            )

        if number is not None and not number > 0:
            raise ValueError(
                "Partition number must be a positive integer."
            )

        return number

    @options.add
    @Argument("--bad", bad=True)
    def bad(self, bad=False):
        """Set the partition as unbootable"""
        return bad

    @options.add
    @Argument("--oneshot", oneshot=True)
    def oneshot(self, oneshot=False):
        """Set the partition to be tried once"""
        return oneshot

    def __call__(self):
        if self.bad == False:
            try:
                self.partition.tries = 1
            except subprocess.CalledProcessError as err:
                raise CommandExit(
                    "Failed to set remaining tries for partition '{}'."
                    .format(self.partition)
                ) from err

            if self.oneshot == False:
                try:
                    self.partition.successful = 1
                except subprocess.CalledProcessError as err:
                    raise CommandExit(
                        "Failed to set success flag for partition '{}'."
                        .format(self.partition)
                    ) from err

                self.logger.warning(
                    "Set partition '{}' as successfully booted."
                    .format(self.partition)
                )

            else:
                try:
                    self.partition.successful = 0
                except subprocess.CalledProcessError as err:
                    raise CommandExit(
                        "Failed to unset successful flag for partition '{}'."
                        .format(self.partition)
                    ) from err

                self.logger.warning(
                    "Set partition '{}' as not yet successfully booted."
                    .format(self.partition)
                )

            try:
                self.partition.prioritize()
            except subprocess.CalledProcessError as err:
                raise CommandExit(
                    "Failed to prioritize partition '{}'."
                    .format(self.partition)
                ) from err

            self.logger.info(
                "Set partition '{}' as the highest-priority bootable part."
                .format(self.partition)
            )

        else:
            try:
                self.partition.attribute = 0x000
            except subprocess.CalledProcessError as err:
                raise CommandExit(
                    "Failed to zero attributes for partition '{}'."
                    .format(self.partition)
                ) from err

            self.logger.warning(
                "Set partition '{}' as a zero-priority unbootable part."
                .format(self.partition)
            )

    global_options = depthchargectl.global_options
    config_options = depthchargectl.config_options



#! /usr/bin/env python3
# SPDX-License-Identifier: GPL-2.0-or-later

# depthcharge-tools depthchargectl build subcommand
# Copyright (C) 2020-2023 Alper Nebi Yasak <alpernebiyasak@gmail.com>
# See COPYRIGHT and LICENSE files for full copyright information.

import argparse
import collections
import configparser
import logging
import os
import shlex
import textwrap

from pathlib import Path
from functools import lru_cache

from depthcharge_tools import __version__
from depthcharge_tools.mkdepthcharge import mkdepthcharge
from depthcharge_tools.utils.argparse import (
    Command,
    Argument,
    Group,
    CommandExit,
)
from depthcharge_tools.utils.os import (
    Partition,
)
from depthcharge_tools.utils.pathlib import (
    copy,
)
from depthcharge_tools.utils.platform import (
    KernelEntry,
    cpu_microcode,
    vboot_keys,
    installed_kernels,
    root_requires_initramfs,
)
from depthcharge_tools.utils.subprocess import (
    fdtget,
)

from depthcharge_tools.depthchargectl import depthchargectl


class SizeTooBigError(CommandExit):
    def __init__(self):
        super().__init__(
            "Couldn't build a small enough image for this board.",
            returncode=4,
        )


class InitramfsSizeTooBigError(SizeTooBigError):
    def __init__(self):
        super(SizeTooBigError, self).__init__(
            "Couldn't build a small enough image for this board. "
            "This is usually solvable by making the initramfs smaller, "
            "check your OS's documentation on how to do so.",
            returncode=3,
        )


@depthchargectl.subcommand("build")
class depthchargectl_build(
    depthchargectl,
    prog="depthchargectl build",
    usage="%(prog)s [options] [KERNEL_VERSION]",
    add_help=False,
):
    """Buld a depthcharge image for the running system."""

    _logger = depthchargectl._logger.getChild("build")
    config_section = "depthchargectl/build"

    @Group
    def positionals(self):
        """Positional arguments"""

    @positionals.add
    @Argument
    def kernel_version(self, kernel_version=None):
        """Installed kernel version to build an image for."""
        if isinstance(kernel_version, KernelEntry):
            return kernel_version

        kernels = installed_kernels(
            root=self.root_mountpoint,
            boot=self.boot_mountpoint,
        )

        kernel_arches = self.board.arch.kernel_arches
        for k in list(kernels):
            if k.arch not in kernel_arches:
                self.logger.info(
                    "Ignoring kernel '{}' incompatible with board arch."
                    .format(k.release or "(unknown)")
                )
                kernels.remove(k)

        if isinstance(kernel_version, str):
            kernel = max(
                (k for k in kernels if k.release == kernel_version),
                default=None,
            )

            if kernel is None:
                raise ValueError(
                    "Could not find an installed kernel for version '{}'."
                    .format(kernel_version)
                )

        elif kernels:
            kernel = max(kernels)

        else:
            self.logger.warning(
                "Could not find any installed kernel."
            )
            kernel = None

        return kernel

    @Group
    def options(self):
        """Options"""

    @depthchargectl.board.copy()
    def board(self, codename=""):
        board = super().board

        if board is None:
            raise ValueError(
                "Cannot build depthcharge images when no board is specified.",
            )

        return board

    @depthchargectl.zimage_initramfs_hack.copy()
    def zimage_initramfs_hack(self, hack=None):
        hack = super().zimage_initramfs_hack

        if hack not in (None, "set-init-size", "pad-vmlinuz"):
            raise ValueError(
                "Unknown zimage initramfs support hack '{}'."
                .format(hack)
            )

        return hack

    @Group
    def custom_kernel_options(self):
        """Custom kernel specification"""

    @custom_kernel_options.add
    @Argument("--kernel-release", nargs=1)
    def kernel_release(self, name=None):
        """Release name for the kernel used in image name"""
        if name is None and self.kernel_version is not None:
            if self.kernel == self.kernel_version.kernel:
                name = self.kernel_version.release

        return name

    @custom_kernel_options.add
    @Argument("--kernel", nargs=1)
    def kernel(self, file_=None):
        """Kernel executable"""
        if file_ is None and self.kernel_version is not None:
            file_ = self.kernel_version.kernel

        # vmlinuz is always mandatory
        if file_ is None and self.kernel_release is not None:
            raise ValueError(
                "No vmlinuz file found for version '{}'."
                .format(self.kernel_release)
            )

        elif file_ is None:
            raise ValueError("No vmlinuz file found.")

        return Path(file_)

    @custom_kernel_options.add
    @Argument("--initramfs", nargs='+')
    def initrd(self, *files):
        """Ramdisk images"""
        # Trigger more important errors first
        self.kernel

        if not files and self.kernel_version is not None:
            if self.kernel == self.kernel_version.kernel:
                microcode = cpu_microcode(self.boot_mountpoint)
                files = [*microcode, self.kernel_version.initrd]

        if self.ignore_initramfs:
            for file in files:
                self.logger.warning(
                    "Ignoring initramfs '{}' as configured."
                    .format(file)
                )
            return None

        if len(files) == 1 and files[0] in (None, "None", "none"):
            self.logger.warning("Not using initramfs.")
            return None

        # Initramfs is optional.
        if not files and self.kernel_release is not None:
            self.logger.info(
                "No initramfs file found for version '{}'."
                .format(self.kernel_release)
            )
            return None

        elif not files:
            self.logger.info("No initramfs file found.")
            return None

        else:
            return [Path(file) for file in files]

    @custom_kernel_options.add
    @Argument("--fdtdir", nargs=1)
    def fdtdir(self, dir_=None):
        """Directory to search device-tree binaries for the board"""
        if dir_ is None and self.kernel_version is not None:
            if self.kernel == self.kernel_version.kernel:
                dir_ = self.kernel_version.fdtdir

        if dir_ is None:
            return None
        else:
            return Path(dir_)

    @custom_kernel_options.add
    @Argument("--dtbs", nargs="+", metavar="FILE")
    def dtbs(self, *files):
        """Device-tree binary files to use instead of searching fdtdir"""
        # Trigger more important errors first
        self.kernel

        # Device trees are optional based on board configuration.
        if self.board.dt_compatible and len(files) == 0:
            if self.fdtdir is None and self.kernel_release is not None:
                raise ValueError(
                    "No dtb directory found for version '{}', "
                    "but this board needs a dtb."
                    .format(self.kernel_release)
                )

            elif self.fdtdir is None:
                raise ValueError(
                    "No dtb directory found, "
                    "but this board needs a dtb."
                )

            self.logger.info(
                "Searching '{}' for dtbs compatible with pattern '{}'."
                .format(self.fdtdir, self.board.dt_compatible.pattern)
            )

            def is_compatible(dt_file):
                return any(
                    self.board.dt_compatible.fullmatch(compat)
                    for compat in fdtget.get(
                        dt_file, "/", "compatible", default="",
                    ).split()
                )

            files = list(filter(
                is_compatible,
                self.fdtdir.glob("**/*.dtb"),
            ))

            if len(files) == 0:
                raise ValueError(
                    "No dtb file compatible with pattern '{}' found in '{}'."
                    .format(self.board.dt_compatible.pattern, self.fdtdir)
                )

        else:
            files = [Path(f) for f in files]

        if self.board.image_format == "zimage" and len(files) != 0:
            raise ValueError(
                "Image format '{}' doesn't support dtb files."
                .format(self.board.image_format)
            )

        return sorted(files, key=lambda f: f.name)

    @options.add
    @Argument("--description", nargs=1)
    def description(self, desc=None):
        """Human-readable description for the image"""
        if desc is None and self.kernel_version is not None:
            desc = self.kernel_version.description

        return desc

    @options.add
    @depthchargectl.root.copy("--root")
    def root(self, root=None):
        """Root device to add to the kernel cmdline"""
        if root in ("", "None", "none"):
            return None

        root = super().root

        if isinstance(root, Path):
            mnt = self.root_mountpoint
            root = self.diskinfo.by_mountpoint("/", fstab_only=True)

            if root:
                self.logger.info(
                    "Using root '{}' set in '{}'."
                    .format(root, mnt / "etc" / "fstab")
                )
                return root

            dev = self.diskinfo.by_mountpoint(mnt)
            uuid = self.diskinfo.get_uuid(dev)
            if uuid:
                root = "UUID={}".format(uuid.upper())
                self.logger.warning(
                    "Using '{}' as root from currently mounted '{}'."
                    .format(root, dev)
                )
                return root

            if mnt != Path("/").resolve():
                raise ValueError(
                    "Couldn't convert mountpoint '{}' to a root cmdline."
                    .format(mnt)
                )

            raise ValueError(
                "Couldn't figure out a root cmdline for this system."
                .format(mnt)
            )

        if not self.diskinfo.evaluate(root):
            self.logger.warning(
                "Using root '{}' but it's not a device or mountpoint."
                .format(root)
            )

        return str(root)

    @depthchargectl.kernel_cmdline.copy()
    def kernel_cmdline(self, *cmds):
        # This is some deep magic that evaluates self.kernel_cmdline
        # according to the parent definition and sets it in self.
        cmdline = super().kernel_cmdline

        # This evaluates self.root with the above self.kernel_cmdline,
        # then continues here to set self.kernel_cmdline a second time.
        append_root = self.root is not None

        for c in list(cmdline):
            lhs, _, rhs = c.partition("=")
            if lhs.lower() != "root":
                continue

            if rhs == self.root:
                append_root = False
                continue

            if self.root is None:
                self.logger.warning(
                    "Kernel cmdline has a root '{}', keeping it."
                    .format(rhs)
                )
                continue

            self.logger.warning(
                "Kernel cmdline has a different root '{}', removing it."
                .format(rhs)
            )
            cmdline.remove(c)

        if append_root:
            self.logger.info(
                "Appending 'root={}' to kernel cmdline."
                .format(self.root)
            )
            cmdline.append('root={}'.format(self.root))

        if self.ignore_initramfs:
            self.logger.warning(
                "Ignoring initramfs as configured, "
                "appending 'noinitrd' to the kernel cmdline."
                .format(self.initrd)
            )
            cmdline.append("noinitrd")

        # Linux kernel without an initramfs only supports certain
        # types of root parameters, check for them.
        if self.initrd is None and self.root is not None:
            if root_requires_initramfs(self.root):
                raise ValueError(
                    "An initramfs is required for root '{}'."
                    .format(self.root)
                )

        return cmdline

    @options.add
    @Argument("--compress", nargs="+", metavar="TYPE")
    def compress(self, *compress):
        """Compression types to attempt."""

        # Allowed compression levels. We will call mkdepthcharge by
        # hand multiple times for these.
        for c in compress:
            if c not in ("none", "lz4", "lzma"):
                raise ValueError(
                    "Unsupported compression type '{}'."
                    .format(t)
                )

        if len(compress) == 0:
            compress = ["none"]
            if self.board.boots_lz4_kernel:
                compress += ["lz4"]
            if self.board.boots_lzma_kernel:
                compress += ["lzma"]

            # zimage doesn't support compression
            if self.board.image_format == "zimage":
                compress = ["none"]

        return sorted(set(compress), key=compress.index)

    @options.add
    @Argument("--timestamp", nargs=1)
    def timestamp(self, seconds=None):
        """Build timestamp for the image"""
        if seconds is None:
            if "SOURCE_DATE_EPOCH" in os.environ:
                seconds = os.environ["SOURCE_DATE_EPOCH"]

        # Initramfs date is bound to be later than vmlinuz date, so
        # prefer that if possible.
        if seconds is None:
            if self.initrd is not None:
                seconds = max(
                    int(initrd.stat().st_mtime)
                    for initrd in self.initrd
                )
            else:
                seconds = int(self.kernel.stat().st_mtime)

        if seconds is None:
            self.logger.error(
                "Couldn't determine a timestamp from initramfs "
                "nor vmlinuz."
            )

        return seconds

    @options.add
    @Argument("-o", "--output", nargs=1)
    def output(self, path=None):
        """Output image to path instead of storing in images-dir"""
        if path is None:
            image_name = "{}.img".format(self.kernel_release or "default")
            path = self.images_dir / image_name

        return Path(path)

    def __call__(self):
        self.logger.warning(
            "Building depthcharge image for board '{}' ('{}')."
            .format(self.board.name, self.board.codename)
        )

        self.logger.info(
            "Building for kernel version '{}'."
            .format(self.kernel_release or "(unknown)")
        )

        # Images dir might not have been created at install-time
        os.makedirs(self.output.parent, exist_ok=True)

        # Build to a temporary file so we do not overwrite existing
        # images with an unbootable image.
        outtmp = self.tmpdir / "{}.tmp".format(self.output.name)

        # Try to keep output reproducible.
        if self.timestamp is not None:
            os.environ["SOURCE_DATE_EPOCH"] = str(self.timestamp)

        # Error early if initramfs is absolutely too big to fit
        initrd_size = (
            sum(initrd.stat().st_size for initrd in self.initrd)
            if self.initrd is not None
            else 0
        )
        if initrd_size >= self.board.image_max_size:
            self.logger.error(
                "Initramfs alone is larger than the maximum image size."
            )
            raise InitramfsSizeTooBigError()

        # The earliest boards apparently have an off-by-one error while
        # loading the chosen dtb, adding each file twice solves it.
        dtbs = (
            self.dtbs
            if not self.board.loads_dtb_off_by_one else
            [dtb for dtb in self.dtbs for _ in (0, 1)]
        )

        # Skip compress="none" if inputs wouldn't fit max image size
        compress_list = self.compress
        inputs_size = sum([
            self.kernel.stat().st_size,
            initrd_size,
            *(dtb.stat().st_size for dtb in dtbs),
        ])

        if inputs_size > self.board.image_max_size and "none" in compress_list:
            self.logger.info(
                "Inputs are too big, skipping uncompressed build."
            )
            compress_list.remove("none")

        if not compress_list:
            raise SizeTooBigError()

        # Avoid passing format-specific options unrelated to board format
        image_format_opts = {
            "image_format": self.board.image_format,
        }

        if self.board.image_format == "fit":
            image_format_opts["name"] = self.description
            image_format_opts["patch_dtbs"] = not self.board.loads_fit_ramdisk

            if self.board.fit_ramdisk_load_address is not None:
                image_format_opts["ramdisk_load_address"] = (
                    self.board.fit_ramdisk_load_address
                )

        elif self.board.image_format == "zimage":
            if not self.board.loads_zimage_ramdisk:
                hack = self.zimage_initramfs_hack
                image_format_opts["pad_vmlinuz"] = (hack == "pad-vmlinuz")
                image_format_opts["set_init_size"] = (hack == "set-init-size")

        for compress in compress_list:
            self.logger.info("Trying with compression '{}'.".format(compress))
            tmpdir = self.tmpdir / "mkdepthcharge-{}".format(compress)

            try:
                mkdepthcharge(
                    arch=self.board.arch,
                    cmdline=self.kernel_cmdline,
                    compress=compress,
                    dtbs=dtbs,
                    **image_format_opts,
                    kernel_start=self.board.image_start_address,
                    initramfs=self.initrd,
                    keyblock=self.vboot_keyblock,
                    output=outtmp,
                    signprivate=self.vboot_private_key,
                    signpubkey=self.vboot_public_key,
                    vmlinuz=self.kernel,
                    tmpdir=tmpdir,
                    verbosity=self.verbosity,
                )

            except Exception as err:
                raise CommandExit(
                    "Failed while creating depthcharge image.",
                ) from err

            if outtmp.stat().st_size < self.board.image_max_size:
                break

            self.logger.warning(
                "Image with compression '{}' is too big for this board."
                .format(compress)
            )

        else:
            # The necessary zimage padding might be too big, actually
            # check if reducing the initramfs would make things fit.
            if outtmp.stat().st_size - initrd_size < self.board.image_max_size:
                raise InitramfsSizeTooBigError()
            else:
                raise SizeTooBigError()

        self.logger.info("Copying newly built image to output.")
        try:
            copy(outtmp, self.output)
        except PermissionError as err:
            raise PermissionError(
                "Couldn't copy to '{}', permission denied."
                .format(self.output)
            )

        self.logger.warning(
            "Built depthcharge image for kernel version '{}'."
            .format(self.kernel_release or "(unknown)")
        )
        return self.output

    global_options = depthchargectl.global_options
    config_options = depthchargectl.config_options




# SPDX-License-Identifier: GPL-2.0-or-later

# depthcharge-tools python package setup script
# Copyright (C) 2020-2023 Alper Nebi Yasak <alpernebiyasak@gmail.com>
# See COPYRIGHT and LICENSE files for full copyright information.

#! /usr/bin/env python3

import pathlib
import setuptools

root = pathlib.Path(__file__).resolve().parent
readme = (root / 'README.rst').read_text()

setuptools.setup(
    name='depthcharge-tools',
    version='0.6.2',
    description='Tools to manage the Chrome OS bootloader',
    long_description=readme,
    long_description_content_type="text/x-rst",
    url='https://github.com/alpernebbi/depthcharge-tools',
    author='Alper Nebi Yasak',
    author_email='alpernebiyasak@gmail.com',
    license='GPL2+',
    license_files=["LICENSE", "COPYRIGHT"],
    classifiers=[
        'Development Status :: 3 - Alpha',
        'Environment :: Console',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: GNU General Public License v2 or later (GPLv2+)',
        'Operating System :: POSIX :: Linux',
        'Programming Language :: Python :: 3',
        'Topic :: System :: Boot',
    ],
    entry_points={
        'console_scripts': [
            'mkdepthcharge=depthcharge_tools.mkdepthcharge:mkdepthcharge.main',
            'depthchargectl=depthcharge_tools.depthchargectl:depthchargectl.main',
        ],
    },
    keywords='ChromeOS ChromiumOS depthcharge vboot vbutil_kernel',
    packages=setuptools.find_packages(),
    package_data={
        "depthcharge_tools": ["config.ini", "boards.ini"],
    },
    install_requires=[
        'setuptools',
    ],
)



#! /usr/bin/env python3
# SPDX-License-Identifier: GPL-2.0-or-later

# depthcharge-tools depthchargectl check subcommand
# Copyright (C) 2020-2023 Alper Nebi Yasak <alpernebiyasak@gmail.com>
# See COPYRIGHT and LICENSE files for full copyright information.

import argparse
import logging

from pathlib import Path

from depthcharge_tools import __version__
from depthcharge_tools.utils.argparse import (
    Command,
    Argument,
    Group,
    CommandExit,
)
from depthcharge_tools.utils.subprocess import (
    fdtget,
    vbutil_kernel,
)

from depthcharge_tools.depthchargectl import depthchargectl


class SizeTooBigError(CommandExit):
    def __init__(self, image, image_size, max_size):
        message = (
            "Image '{}' ({} bytes) must be smaller than {} bytes."
            .format(image, image_size, max_size)
        )

        self.image = image
        self.image_size = image_size
        self.max_size = max_size
        super().__init__(output=False, returncode=3, message=message)


class NotADepthchargeImageError(CommandExit):
    def __init__(self, image):
        message = (
            "Image '{}' is not a depthcharge image."
            .format(image)
        )

        self.image = image
        super().__init__(output=False, returncode=4, message=message)


class VbootSignatureError(CommandExit):
    def __init__(self, image):
        message = (
            "Depthcharge image '{}' is not signed by the configured keys."
            .format(image)
        )

        self.image = image
        super().__init__(output=False, returncode=5, message=message)


class ImageFormatError(CommandExit):
    def __init__(self, image, board_format):
        message = (
            "Image '{}' must be in '{}' format."
            .format(image, board_format)
        )

        self.image = image
        self.board_format = board_format
        super().__init__(output=False, returncode=6, message=message)


class MissingDTBError(CommandExit):
    def __init__(self, image, compat):
        message = (
            "Image '{}' must have a device-tree binary compatible with pattern '{}'."
            .format(image, compat)
        )

        self.image = image
        self.compat = compat
        super().__init__(output=False, returncode=7, message=message)


@depthchargectl.subcommand("check")
class depthchargectl_check(
    depthchargectl,
    prog="depthchargectl check",
    usage = "%(prog)s [options] IMAGE",
    add_help=False,
):
    """Check if a depthcharge image can be booted."""

    _logger = depthchargectl._logger.getChild("check")
    config_section = "depthchargectl/check"

    @Group
    def positionals(self):
        """Positional arguments"""

    @positionals.add
    @Argument
    def image(self, image):
        """Depthcharge image to check validity of."""
        image = Path(image)

        if not image.is_file():
            raise ValueError("Image argument must be a file")

        return image

    @depthchargectl.board.copy()
    def board(self, codename=""):
        """Assume we're running on the specified board"""
        board = super().board

        if board is None:
            raise ValueError(
                "Cannot check depthcharge images when no board is specified.",
            )

        return board

    @depthchargectl.zimage_initramfs_hack.copy()
    def zimage_initramfs_hack(self, hack=None):
        hack = super().zimage_initramfs_hack

        if hack not in (None, "set-init-size", "pad-vmlinuz"):
            raise ValueError(
                "Unknown zimage initramfs support hack '{}'."
                .format(hack)
            )

        return hack


    def __call__(self):
        image = self.image

        self.logger.warning(
            "Verifying depthcharge image for board '{}' ('{}')."
            .format(self.board.name, self.board.codename)
        )

        self.logger.info("Checking if image fits into size limit.")
        image_size = image.stat().st_size
        if image_size > self.board.image_max_size:
            raise SizeTooBigError(
                image,
                image_size,
                self.board.image_max_size,
            )

        self.logger.info("Checking depthcharge image validity.")
        if vbutil_kernel(
            "--verify", image,
            check=False,
        ).returncode != 0:
            raise NotADepthchargeImageError(image)

        self.logger.info("Checking depthcharge image signatures.")
        if self.vboot_public_key is not None:
            if vbutil_kernel(
                "--verify", image,
                "--signpubkey", self.vboot_public_key,
                check=False,
            ).returncode != 0:
                raise VbootSignatureError(image)

        itb = self.tmpdir / "{}.itb".format(image.name)
        vbutil_kernel(
            "--get-vmlinuz", image,
            "--vmlinuz-out", itb,
            check=False,
        )

        if self.board.image_format == "fit":
            self.logger.info("Checking FIT image format.")
            nodes = fdtget.subnodes(itb)
            if "images" not in nodes and "configurations" not in nodes:
                raise ImageFormatError(image, self.board.image_format)

            def is_compatible(dt_file, conf_path):
                return any(
                    self.board.dt_compatible.fullmatch(compat)
                    for compat in fdtget.get(
                        dt_file, conf_path, "compatible", default="",
                    ).split()
                )

            self.logger.info("Checking included DTB binaries.")
            for conf in fdtget.subnodes(itb, "/configurations"):
                conf_path = "/configurations/{}".format(conf)
                if is_compatible(itb, conf_path):
                    break

                dtb = fdtget.get(itb, conf_path, "fdt")
                dtb_path = "/images/{}".format(dtb)
                dtb_data = fdtget.get(itb, dtb_path, "data", type=bytes)
                dtb_file = self.tmpdir / "{}.dtb".format(conf)
                dtb_file.write_bytes(dtb_data)

                if is_compatible(dtb_file, "/"):
                    break
            else:
                raise MissingDTBError(
                    image, self.board.dt_compatible.pattern,
                )

        self.logger.warning(
            "This command is incomplete, the image might be unbootable "
            "despite passing currently implemented checks."
        )

    global_options = depthchargectl.global_options
    config_options = depthchargectl.config_options
